from kamodo import Kamodo, kamodofy
import pandas as pd
import numpy as np
import scipy
import time
import datetime
from datetime import timezone
import urllib, json
import plotly.express as px
import plotly.graph_objects as go
import pandas as pd
from pandas import DatetimeIndex
from collections.abc import Iterable


def ror_get_info(runID):
    '''Query run information for given runID'''
    server = "https://ccmc.gsfc.nasa.gov/RoR_WWW/VMR"
    query = '{}/files.php?id={}'.format(server, runID)
    response = urllib.request.urlopen(query)
    data = json.loads(response.read())
    return data

def ror_show_info(runID):
    '''Display run information for a given runID.'''
    result=ror_get_info(runID)
    print("Run information for runID =",runID,"from the CCMC RoR system.")
    for item in result['info']:
        for k in item:
            print(k.rjust(25),':',item[k])
    sats = []
    for sat in result['satellites']:
        satname = sat['name']
        sats.append(satname)
    k='Satellite extractions'
    print(k.rjust(25),':',sats)

def ror_return_satellites(runID):
    '''Display list of satellites as python array for given runID.'''
    result=ror_get_info(runID)
    sats = []
    for sat in result['satellites']:
        satname = sat['name']
        sats.append(satname)
    return sats

def ror_get_extraction(runID, coord, satellite):
    '''Query for file contents from server'''
    server = "https://ccmc.gsfc.nasa.gov/RoR_WWW/VMR"
    query = '{}/{}/{}/{}_{}.txt'.format(server, runID, satellite, coord, satellite)
    # print(query)
    response = urllib.request.urlopen(query)
    file = response.read()
    return file

class SATEXTRACT(Kamodo):
    def __init__(self,
            runID, # ccmc runs-on-request run id
            coord, # coordinate system
            satellite, # satellite
            debug=0,
            server="https://ccmc.gsfc.nasa.gov/RoR_WWW/VMR",
            **kwargs):
        super(SATEXTRACT, self).__init__(**kwargs)
        self.verbose=False # overrides kwarg
        # self.symbol_registry=dict() # wipes any user-defined symbols
        # self.signatures=dict() # wipes any user-defined signatures
        self.RE=6.3781E3
        self.server = server # should be set by keyword
        self.runID = runID
        self.coordinates = coord
        self.satellite = satellite
        self.debug = debug
        if self.debug > 0:
            print(' -server: CCMC RoR')
            print(' -runID: ',runID)
            print(' -coordinate system: ',coord)
            print(' -satellite: ',satellite)
        self.variables=dict()
        self.file = ror_get_extraction(runID, coord, satellite).decode('ascii')
        self.parse_file()
        ts=self.tsarray[0]
        self.start = datetime.datetime.fromtimestamp(ts,tz=timezone.utc).strftime("%Y-%m-%dT%H:%M:%SZ")
        ts=self.tsarray[-1]
        self.stop  = datetime.datetime.fromtimestamp(ts,tz=timezone.utc).strftime("%Y-%m-%dT%H:%M:%SZ")
        if self.debug > 0:
            print(" ")
            print(" -date start: ",self.start)
            print("         end: ",self.stop)

        for varname in self.variables:
            if varname == "N":
                continue
            units = self.variables[varname]['units']
            if self.debug > 0:
                print('... registering ',varname,units)
            self.register_variable(varname, units)
        
        # classification of position into coordinates to assist visualizion
        self.possible_coords=('TOD','J2K','GEO','GM','GSM','GSE','SM')
        self.possible_directions=('x','y','z')
        self.coords=dict()
        for varname in self.variables:
            size = self.variables[varname]['size']
            if size == 1:
                # Look for position values
                direction = varname.lower()
                key = self.coordinates
                if key in self.possible_coords and direction in self.possible_directions:
                    if key not in self.coords:
                        self.coords[key] = dict(coord=key)
                    self.coords[key]['size'] = size
                    self.coords[key][direction] = varname

        # Change 'fill' values in data to NaNs
        self.fill2nan()
        
    def parse_file(self):
        import re
        vars=[]
        units=[]
        times=[]
        arrays = []
        if self.debug > 0:
            print("===> Printing File Header ...")
        for line in self.file.splitlines(False):
            A = re.match('^# ', line)
            B = re.match('# Run', line)
            C = re.match('# Coordinate', line)
            D = re.match('# Satellite', line)
            E = re.match('# Year', line)
            F = re.match('# \[year\]', line)
            G = re.match('# Data type', line)
            if A or B or C or D or E or F or G:
                if A:
                    if self.debug > 0:
                        print("-> ",line)
                if B:
                    # Find runname and fill value
                    parts=re.sub(' +', ' ', line).split(' ')
                    self.runname = parts[3]
                    self.fillvalue = parts[6]
                if C:
                    # Check that coordinate system matches
                    parts=re.sub(' +', ' ', line).split(' ')
                    if self.coordinates != parts[3]:
                        print("ERROR: Coordinate system does not match.",self.coordinates,parts[3])
                if D:
                    # Check that satellite name matches
                    parts=re.sub(' +', ' ', line).split(' ')
                    if self.satellite != parts[3]:
                        print("ERROR: Satellite does not match.",self.satellite,parts[3])
                if E:
                    # Variable names, remove . and change N and B_1
                    parts=re.sub(' +', ' ', line).strip().split(' ')
                    for p in parts[7:]:
                        p=re.sub("\.","",p)
                        p=re.sub("B_1","B1",p)
                        p=re.sub("^N$","rho",p)
                        vars.append(p)
                    if self.debug > 1:
                        print(len(vars), vars)
                if F:
                    # Variable units, remove [] and fix exponents
                    parts=re.sub(' +', ' ', line).strip().split(' ')
                    if self.modelname == "GUMICS":
                        # missing x,y,z  --put before other units
                        units.append('R_E')
                        units.append('R_E')
                        units.append('R_E')
                    for p in parts[7:]:
                        if self.modelname == "BATSRUS":
                            # need a unit applied for status variable, currently missing in some
                            if len(units) != len(vars):
                                if vars[len(units)] == "status":
                                    units.append('')
                        p=re.sub("cm\^-3","1/cm^3",p)
                        p=re.sub("m2","m^2",p)
                        p=re.sub("m3","m^3",p)
                        p=re.sub("\[","",p)
                        p=re.sub("\]","",p)
                        p=re.sub("Vm/A","V/A",p) # This is wrong but use it for now
                        p=re.sub("nJ/m","J/m",p) # This is wrong but use it for now
                        units.append(p)
                        if self.modelname == "LFM" and p == "nPa":
                            # missing X,Y,Z,Vol  --unknown Vol units
                            units.append('R_E')
                            units.append('R_E')
                            units.append('R_E')
                            units.append('')
                        if self.modelname == "OpenGGCM" and p == "V/A":
                            # missing eflx,efly,eflz  --units are unknown
                            units.append('')
                            units.append('')
                            units.append('')
                    if self.debug > 1:
                        print(len(units), units)
                if G:
                    # Pull out the model name
                    parts=re.sub(' +', ' ', line).split(' ')
                    self.modelname = parts[3]
            else:
                parts=re.sub(' +', ' ', line).strip().split(' ')
                year=parts[0]
                month=parts[1]
                day=parts[2]
                hour=parts[3]
                minute=parts[4]
                second=parts[5]
                ms=0
                if '.' in second:
                    (second,ms)=second.split('.')
                dd=datetime.datetime(int(year),int(month),int(day),
                                     hour=int(hour),minute=int(minute),second=int(second),
                                     microsecond=int(ms)*1000,tzinfo=datetime.timezone.utc)
                times.append(dd)
                for s in parts[6:]:
                    arrays.append(float(s))
        # self.dtarray=np.array([dd for dd in times])
        self.dtarray = pd.to_datetime(times)
        self.tsarray = np.array([d.timestamp() for d in self.dtarray])
        self.dtarrayclean = np.array([datetime.datetime.fromtimestamp(d,tz=timezone.utc).strftime("%Y-%m-%d %H:%M:%S") for d in self.tsarray])
        
        nvar=len(vars)
        nval=len(arrays)
        npos=int(nval/nvar)
        arrays=np.array(arrays)
        arrays=arrays.reshape((npos,nvar))

        i=0
        for var in vars:
            self.variables[var] = dict(units=units[i],
                                       data=arrays[:,i],
                                       size=1,
                                       fill=self.fillvalue)
            i+=1
        
        return
                
    def register_variable(self, varname, units):
        """register variables into Kamodo for this service, CCMC ROR satellite extractions"""

        data =  self.variables[varname]['data']
        times = self.dtarray

        ser = pd.Series(data, index=pd.DatetimeIndex(times))


        @kamodofy(units = units, 
                  citation = "De Zeeuw 2020",
                  data = None)
        def interpolate(t=times):
            ts = t
            isiterable = isinstance(t, Iterable)

            if isinstance(ts, DatetimeIndex):
                pass
            elif isinstance(ts, float):
                ts = pd.to_datetime([ts], utc=True, unit='s')
            elif isiterable:
                if isinstance(ts[0], float):
                    ts = pd.to_datetime(ts, utc=True, unit='s')
                ts = DatetimeIndex(ts)
            else:
                raise NotImplementedError(ts)
            
            ser_ = ser.reindex(ser.index.union(ts))
            ser_interpolated = ser_.interpolate(method='time')
            result = ser_interpolated.reindex(ts)
            if isiterable:
                return result.values
            else:
                return result.values[0]

        # store the interpolator
        self.variables[varname]['interpolator'] = interpolate

        # update docstring for this variable
        interpolate.__doc__ = "A function that returns {} in [{}].".format(varname,units)

        var_reg = '{}__{}'.format(varname, self.coordinates)
        self[var_reg] = interpolate

    def fill2nan(self):
        '''
        Replaces fill value in data with NaN.
        Not Yet called by default. Call as needed.
        '''
        for varname in self.variables:
            data = self.variables[varname]['data']
            fill = self.variables[varname]['fill']
            if fill is not None:
                mask = data==float(fill)
                nbad = np.count_nonzero(mask)
                if nbad > 0:
                    if self.debug > 0:
                        print("Found",nbad,"fill values, replacing with NaN for variable",
                              varname,"of size",data.size)
                data[mask]=np.nan
                self.variables[varname]['data'] = data

    def info(self):
        '''
        Show information stored in this Kamodo object.
        '''
        print('Kamodo object information:')
        print('  server         CCMC RoR')
        print('  runID         ',self.runID)
        print('  coordinates   ',self.coordinates)
        print('  satellite     ',self.satellite)
        print('  run start     ',self.start)
        print('  run end       ',self.stop)
        print('  variables')
        for varname in self.variables:
            units = self.variables[varname]['units']
            print('     ',varname,' [',units,']')

    def get_plot(self, type="1Dpos", scale="R_E", var="", groupby="all",
                 quiver=False, quiverscale="5.", quiverskip="0"):
        '''
        Return a plotly figure object.
        type = 1Dvar => 1D plot of variable value vs Time (also all variables option)
               1Dpos (default) => 1D location x,y,z vs Time
               3Dpos => 3D location colored by altitude
               3Dvar => View variable on 3D position (also quiver and groupby options)
        scale = km, R_E (default)
        var = variable name for variable value plots
        groupby = day, hour, all (default) => groupings for 3Dvar plots
        quiver = True, False (default) => if var is a vector value and 3Dvar plot, then
               turn on quivers and color by vector magnitude
        quiverscale = 5. (default) => length of quivers in units of RE
        quiverskip =  (default) => now many quivers to skip displaying
        '''
        quiverscale=float(quiverscale)
        if scale == "km":
            quiverscale=quiverscale*self.RE
        quiverskip=int(quiverskip)
        coord=self.coordinates
        
        # Set plot title for plots
        txttop=self.satellite + " position extracted from run " + self.runname + "<br>"\
            + self.start + " to " + self.stop + "<br>" + coord
        
        if type == '1Dvar':
            if var == "":
                print("No plot variable passed in.")
                return
            fig=go.Figure()
            if var == "all":
                # Create menu pulldown for each variable
                steps = []
                i=0
                for varname in self.variables:
                    ytitle=varname+" ["+self.variables[varname]['units']+"]"
                    step = dict(
                        label=ytitle,
                        method="update",
                        args=[{"visible": [False] * len(self.variables)}],
                    )
                    step["args"][0]["visible"][i] = True  # Toggle i'th trace to "visible"
                    steps.append(step)
                    if self.variables[varname]['size'] == 1:
                        x=self.variables[varname]['data']
                        fig.add_trace(go.Scatter(x=self.dtarray, y=x, mode='lines+markers',
                                                 name=varname, visible=False))
                    elif self.variables[varname]['size'] == 3:
                        x=self.variables[varname]['data'][:,0]
                        y=self.variables[varname]['data'][:,1]
                        z=self.variables[varname]['data'][:,2]
                        fig.add_trace(go.Scatter(x=self.dtarray, y=x, mode='lines+markers',
                                                 name=varname, visible=False))
                        fig.add_trace(go.Scatter(x=self.dtarray, y=y, mode='lines+markers',
                                                 name=varname, visible=False))
                        fig.add_trace(go.Scatter(x=self.dtarray, y=z, mode='lines+markers',
                                                 name=varname, visible=False))
                    i+=1
                fig.data[0].visible=True
                fig.update_layout(updatemenus=list([dict(buttons=steps)]))
                fig.update_xaxes(title_text="Time")
                fig.update_layout(hovermode="x")
                fig.update_layout(title_text=txttop)
                
            else:
                # Standard single/vector variable plot
                varn=var
                if var in self.variables:
                    size=self.variables[var]['size']
                else:
                    if var+'_x' in self.variables:
                        size=0
                    else:
                        print("Invalid plot variable passed in.")
                        return
                if size == 0:
                    var=varn+'_x'
                    x=self.variables[var]['data']
                    fig.add_trace(go.Scatter(x=self.dtarray, y=x, mode='lines+markers', name=var))
                    var=varn+'_y'
                    x=self.variables[var]['data']
                    fig.add_trace(go.Scatter(x=self.dtarray, y=x, mode='lines+markers', name=var))
                    var=varn+'_z'
                    x=self.variables[var]['data']
                    fig.add_trace(go.Scatter(x=self.dtarray, y=x, mode='lines+markers', name=var))
                elif size == 1:
                    x=self.variables[var]['data']
                    fig.add_trace(go.Scatter(x=self.dtarray, y=x, mode='lines+markers', name=var))
                elif size == 3:
                    x=self.variables[var]['data'][:,0]
                    y=self.variables[var]['data'][:,1]
                    z=self.variables[var]['data'][:,2]
                    fig.add_trace(go.Scatter(x=self.dtarray, y=x, mode='lines+markers', name=var))
                    fig.add_trace(go.Scatter(x=self.dtarray, y=y, mode='lines+markers', name=var))
                    fig.add_trace(go.Scatter(x=self.dtarray, y=z, mode='lines+markers', name=var))
                ytitle=varn+" ["+self.variables[var]['units']+"]"
                fig.update_xaxes(title_text="Time")
                fig.update_yaxes(title_text=ytitle)
                fig.update_layout(hovermode="x")
                fig.update_layout(title_text=txttop)
 
            return fig
        
        if type == '1Dpos':
            fig=go.Figure()
            xvarname = self.coords[coord]['x']
            if self.coords[coord]['size'] == 1:
                x=self.variables[self.coords[coord]['x']]['data']
                y=self.variables[self.coords[coord]['y']]['data']
                z=self.variables[self.coords[coord]['z']]['data']
            elif self.coords[coord]['size'] == 3:
                x=self.variables[self.coords[coord]['x']]['data'][:,0]
                y=self.variables[self.coords[coord]['y']]['data'][:,1]
                z=self.variables[self.coords[coord]['z']]['data'][:,2]
            if scale == "km":
                if self.variables[xvarname]['units'] == "R_E":
                    x=x*self.RE
                    y=y*self.RE
                    z=z*self.RE
                ytitle="Position [km]"
            else:
                if self.variables[xvarname]['units'] == "km":
                    x=x/self.RE
                    y=y/self.RE
                    z=z/self.RE
                ytitle="Position [R_E]"
            fig.add_trace(go.Scatter(x=self.dtarray, y=x,
                                     mode='lines+markers', name=self.coords[coord]['x']))
            fig.add_trace(go.Scatter(x=self.dtarray, y=y,
                                     mode='lines+markers', name=self.coords[coord]['y']))
            fig.add_trace(go.Scatter(x=self.dtarray, y=z,
                                     mode='lines+markers', name=self.coords[coord]['z']))
            fig.update_xaxes(title_text="Time")
            fig.update_yaxes(title_text=ytitle)
            fig.update_layout(hovermode="x")
            fig.update_layout(title_text=txttop)
 
            return fig
        
        if type == "3Dpos":
            xvarname = self.coords[coord]['x']
            if self.coords[coord]['size'] == 1:
                x=self.variables[self.coords[coord]['x']]['data']
                y=self.variables[self.coords[coord]['y']]['data']
                z=self.variables[self.coords[coord]['z']]['data']
            elif self.coords[coord]['size'] == 3:
                x=self.variables[self.coords[coord]['x']]['data'][:,0]
                y=self.variables[self.coords[coord]['y']]['data'][:,1]
                z=self.variables[self.coords[coord]['z']]['data'][:,2]
            if scale == "km":
                if self.variables[xvarname]['units'] == "R_E":
                    x=x*self.RE
                    y=y*self.RE
                    z=z*self.RE
                r=(np.sqrt(x**2 + y**2 + z**2))-self.RE
            else:
                if self.variables[xvarname]['units'] == "km":
                    x=x/self.RE
                    y=y/self.RE
                    z=z/self.RE
                r=(np.sqrt(x**2 + y**2 + z**2))-1.
            fig=px.scatter_3d(
                x=x,
                y=y,
                z=z,
                color=r)
            bartitle = "Altitude [" + scale + "]"
            fig.update_layout(coloraxis=dict(colorbar=dict(title=bartitle)))
            fig.update_layout(scene=dict(xaxis=dict(title=dict(text="X ["+scale+"]")),
                                         yaxis=dict(title=dict(text="Y ["+scale+"]")),
                                         zaxis=dict(title=dict(text="Z ["+scale+"]"))))
            fig.update_layout(title_text=txttop)
            return fig
        
        if type == "3Dvar":
            if var == "":
                print("No plot variable passed in.")
                return
            xvarname = self.coords[coord]['x']
            vard=self.variables[var]['data']
            varu=self.variables[var]['units']

            if quiver:
                if "_x" in var or "_y" in var or "_z" in var:
                    var=var.split('_')[0]
                    qxvar=var+"_x"
                    qyvar=var+"_y"
                    qzvar=var+"_z"
                    qxvard=self.variables[qxvar]['data']
                    qyvard=self.variables[qyvar]['data']
                    qzvard=self.variables[qzvar]['data']
                    vard = np.sqrt(np.square(qxvard) +\
                                   np.square(qyvard) +\
                                   np.square(qzvard))
                else:
                    print("A vector variable was not passed, turning quiver off.")
                    quiver=False

            cmin=np.amin(vard)
            cmax=np.amax(vard)
            if self.coords[coord]['size'] == 1:
                x=self.variables[self.coords[coord]['x']]['data']
                y=self.variables[self.coords[coord]['y']]['data']
                z=self.variables[self.coords[coord]['z']]['data']
            elif self.coords[coord]['size'] == 3:
                x=self.variables[self.coords[coord]['x']]['data'][:,0]
                y=self.variables[self.coords[coord]['y']]['data'][:,1]
                z=self.variables[self.coords[coord]['z']]['data'][:,2]
            bodyscale=1.    
            if scale == "km":
                if self.variables[xvarname]['units'] == "R_E":
                    x=x*self.RE
                    y=y*self.RE
                    z=z*self.RE
                    bodyscale=self.RE    
                r=(np.sqrt(x**2 + y**2 + z**2))-self.RE
            else:
                if self.variables[xvarname]['units'] == "km":
                    x=x/self.RE
                    y=y/self.RE
                    z=z/self.RE
                r=(np.sqrt(x**2 + y**2 + z**2))-1.
            xmin=np.amin(x)
            xmax=np.amax(x)
            ymin=np.amin(y)
            ymax=np.amax(y)
            zmin=np.amin(z)
            zmax=np.amax(z)
            if quiver:
                tmpx=x+(qxvard*quiverscale/cmax)
                tmpy=y+(qyvard*quiverscale/cmax)
                tmpz=z+(qzvard*quiverscale/cmax)
                xmin=min(xmin,np.amin(tmpx))
                xmax=max(xmax,np.amax(tmpx))
                ymin=min(ymin,np.amin(tmpy))
                ymax=max(ymax,np.amax(tmpy))
                zmin=min(zmin,np.amin(tmpz))
                zmax=max(zmax,np.amax(tmpz))
            
            # Create empty figure to build pieces into
            fig=go.Figure()
            Nplot = 0

            # Start 1 RE sphere, padded to cover all data positions
            dataXYZ = pd.read_csv('https://ccmc.gsfc.nasa.gov/Kamodo/demo/sphereXYZ.csv')
            dataIJK = pd.read_csv('https://ccmc.gsfc.nasa.gov/Kamodo/demo/sphereIJK.csv')
            if scale == "km":
                dataXYZ=dataXYZ*self.RE
            fig.add_mesh3d()
            fig.data[Nplot].x = np.append(dataXYZ['x'],(xmin,xmax))
            fig.data[Nplot].y = np.append(dataXYZ['y'],(ymin,ymax))
            fig.data[Nplot].z = np.append(dataXYZ['z'],(zmin,zmax))
            fig.data[Nplot].i = dataIJK['i']
            fig.data[Nplot].j = dataIJK['j']
            fig.data[Nplot].k = dataIJK['k']
            fig.data[Nplot].facecolor = dataIJK['c']
            fig.data[Nplot].flatshading = True
            fig.data[Nplot].name = '1 R_E sphere'
            fig.data[Nplot].hovertemplate="Earth<extra></extra>"
            fig.data[Nplot].visible=True
            Nplot += 1

            # Array of 'groupby' values and unique values
            if groupby == "day":
                datearray = np.array([datetime.datetime.fromtimestamp(d,tz=timezone.utc).strftime\
                                      ("%Y-%m-%d") for d in self.tsarray])
            elif groupby == "hour":
                datearray = np.array([datetime.datetime.fromtimestamp(d,tz=timezone.utc).strftime\
                                      ("%Y-%m-%d %H") for d in self.tsarray])
            else:
                datearray = np.array([datetime.datetime.fromtimestamp(d,tz=timezone.utc).strftime\
                                      ("all") for d in self.tsarray])
            udates = np.unique(datearray)
            nsteps = len(udates)

            for date in udates:
                # Compute mast to restrict all data in trace
                mask = date == datearray
                # Create figure with data
                if quiver:
                    fig.add_trace(go.Scatter3d(
                        name=date,  x=x[mask], y=y[mask], z=z[mask],
                        mode='markers',
                        marker=dict(
                            size=4,
                            cmin=cmin, cmax=cmax,
                            color=vard[mask],
                            showscale=True,
                            colorscale='Viridis',
                            colorbar=dict(title=var+" ["+varu+"]"),
                        ),
                        customdata=np.vstack((self.dtarrayclean[mask],\
                                              qxvard[mask],\
                                              qyvard[mask],\
                                              qzvard[mask])).T,
                        hovertemplate="<b>"+self.satellite+" Position</b>"+
                        "<br>X: %{x:.4f}<br>Y: %{y:.4f}<br>Z: %{z:.4f}<br>"+
                        qxvar+": %{customdata[1]:.2f}<br>"+
                        qyvar+": %{customdata[2]:.2f}<br>"+
                        qzvar+": %{customdata[3]:.2f}<br>"+
                        " "+var+": %{marker.color:.2f}"+varu+"<br>"+
                        "%{customdata[0]}<br>"+"<extra></extra>",
                        visible=False,
                    ))
                    Nplot += 1
                else:
                    fig.add_trace(go.Scatter3d(
                        name=date,  x=x[mask], y=y[mask], z=z[mask],
                        mode='markers',
                        marker=dict(
                            size=4,
                            cmin=cmin, cmax=cmax,
                            color=vard[mask],
                            showscale=True,
                            colorscale='Viridis',
                            colorbar=dict(title=var+" ["+varu+"]"),
                        ),
                        customdata=self.dtarrayclean[mask],
                        hovertemplate="<b>"+self.satellite+" Position</b>"+
                        "<br>X: %{x:.4f}<br>Y: %{y:.4f}<br>Z: %{z:.4f}<br>"+
                        " "+var+": %{marker.color:.2f}"+varu+"<br>"+
                        "%{customdata}<br>"+"<extra></extra>",
                        visible=False,
                    ))
                    Nplot += 1
            fig.data[1].visible=True

            if quiver:
                for date in udates:
                    # Compute mask to restrict all data in trace
                    mask = date == datearray
                    # Precompute needed values
                    xm=x[mask]
                    ym=y[mask]
                    zm=z[mask]
                    vm=vard[mask]
                    qxm=qxvard[mask]
                    qym=qyvard[mask]
                    qzm=qzvard[mask]
                    xm=xm.reshape(len(xm),1)
                    ym=ym.reshape(len(ym),1)
                    zm=zm.reshape(len(zm),1)
                    vm=vm.reshape(len(vm),1)
                    qxm=qxm.reshape(len(qxm),1)
                    qym=qym.reshape(len(qym),1)
                    qzm=qzm.reshape(len(qzm),1)
                    if quiverskip > 0:
                        for i in range(len(qxm)):
                            if i%(quiverskip+1) > 0:
                                qxm[i]=0.
                                qym[i]=0.
                                qzm[i]=0.
                    xx=np.concatenate((xm,xm+qxm*quiverscale/cmax,xm),axis=1).reshape(3*len(xm))
                    yy=np.concatenate((ym,ym+qym*quiverscale/cmax,ym),axis=1).reshape(3*len(ym))
                    zz=np.concatenate((zm,zm+qzm*quiverscale/cmax,zm),axis=1).reshape(3*len(zm))
                    # Create figure with data
                    fig.add_trace(go.Scatter3d(
                        name=date,
                        x = xx,
                        y = yy,
                        z = zz,
                        mode='lines',
                        line=dict(
                            width=2,
                            color='rgba(22,22,22,0.2)',
                        ),
                        hoverinfo='skip',
                        visible=False,
                    ))
                    Nplot += 1
                fig.data[1+nsteps].visible=True

            # Start selection slider build
            steps = []
            for i in range(nsteps):
                step = dict(
                    method="update",
                    args=[{"visible": [False] * len(fig.data)}],
                    label=udates[i]
                )
                step["args"][0]["visible"][0] = True  # Set first trace to "visible"
                step["args"][0]["visible"][i+1] = True  # Toggle i'th trace to "visible"
                if quiver:
                    step["args"][0]["visible"][i+1+nsteps] = True  # Toggle i'th trace to "visible"
                steps.append(step)

            sliders = [dict(
                active=0,
                currentvalue={"prefix": "Currently showing: "},
                pad={"t": 50},
                steps=steps
            )]
            fig.update_layout(sliders=sliders)

            # Update axis labels and plot title
            fig.update_layout(scene=dict(xaxis=dict(title=dict(text="X ["+scale+"]")),
                                         yaxis=dict(title=dict(text="Y ["+scale+"]")),
                                         zaxis=dict(title=dict(text="Z ["+scale+"]"))))
            fig.update_layout(title_text=txttop)
            fig.update_layout(showlegend=False)
            
            return fig
        

        print('ERROR, reached end of get_plot without any action taken.')
        return

#======
# New class to collect all satellites
#
class SATEXTRACTALL(Kamodo):
    def __init__(self, runID, coord, debug=0, sats=[], **kwargs):
        super(SATEXTRACTALL, self).__init__(**kwargs)
        # Start timer
        tic = time.perf_counter()
        self.verbose=False
        self.symbol_registry=dict()
        self.signatures=dict()
        self.RE=6.3781E3
        self.runID = runID
        self.coordinates = coord
        self.debug = debug
        if self.debug > 0:
            print(' -server: CCMC RoR')
            print(' -runID: ',runID)
            print(' -coordinate system: ',coord)
        self.satellites = dict()
        result=ror_get_info(runID)
        for sat in result['satellites']:
            satname = sat['name']
            if len(sats) == 0 or satname in sats:
                ror = SATEXTRACT(runID, coord, satname, debug=debug)
                self.satellites[satname] = ror
        self.start=ror.start
        self.stop=ror.stop
        self.runname=ror.runname
        self.modelname=ror.modelname
        if self.debug > 0:
            print(' -data extracted from model: ',self.modelname)
        # end timer
        toc = time.perf_counter()
        print(f"Time loading files and registering satellites: {toc - tic:0.4f} seconds")

    def info(self):
        '''
        Show information stored in this Kamodo object.
        '''
        print('Kamodo object information:')
        print('  server         CCMC RoR')
        print('  runID         ',self.runID)
        print('  coordinates   ',self.coordinates)
        print('  satellites')
        for sat in self.satellites:
            print('     ',sat)
        print('  run start     ',self.start)
        print('  run end       ',self.stop)
        print('  variables')
        for sat in self.satellites:
            ror = self.satellites[sat]
            for varname in ror.variables:
                units = ror.variables[varname]['units']
                print('     ',varname,' [',units,']')
            break

    def get_plot(self, type="3Dvar", scale="R_E", var="", groupby="all",
                 quiver=False, quiverscale="5.", quiverskip="0"):
        '''
        Return a plotly figure object.
        type = 3Dvar => View variable on 3D position (also quiver and groupby options)
        scale = km, R_E (default)
        var = variable name for variable value plots
        groupby = day, hour, all (default) => groupings for 3Dvar plots
        quiver = True, False (default) => if var is a vector value and 3Dvar plot, then
               turn on quivers and color by vector magnitude
        quiverscale = 5. (default) => length of quivers in units of RE
        quiverskip =  (default) => now many quivers to skip displaying
        '''
        quiverscale=float(quiverscale)
        if scale == "km":
            quiverscale=quiverscale*self.RE
        quiverskip=int(quiverskip)
        coord=self.coordinates

        # Set plot title for plots
        txttop="Satellite positions extracted from run " + self.runname + "<br>"\
            + self.start + " to " + self.stop + "<br>" + coord

        # set initial values used later
        xmin=0.
        xmax=0.
        ymin=0.
        ymax=0.
        zmin=0.
        zmax=0.
        cmin= 1.e99
        cmax=-1.e99

        if type == "3Dvar":
            if var == "":
                print("No plot variable passed in.")
                return

            # Create empty figure to build pieces into
            fig=go.Figure()
            Nplot = 0

            # Pre-loop to find cmin,cmax to use in all plot pieces
            for sat in self.satellites:
                ror = self.satellites[sat]
                vard=ror.variables[var]['data']

                if quiver:
                    if "_x" in var or "_y" in var or"_z" in var:
                        var2=var.split('_')[0]
                        qxvar=var2+"_x"
                        qyvar=var2+"_y"
                        qzvar=var2+"_z"
                        qxvard=ror.variables[qxvar]['data']
                        qyvard=ror.variables[qyvar]['data']
                        qzvard=ror.variables[qzvar]['data']
                        vard = np.sqrt(np.square(qxvard) +\
                                       np.square(qyvard) +\
                                       np.square(qzvard))
                    else:
                        print("A vector variable was not passed, turning quiver off.")
                        quiver=False

                cmin=min(cmin,np.amin(vard))
                cmax=max(cmax,np.amax(vard))
            
            # Actual plot creation loop
            for sat in self.satellites:
                ror = self.satellites[sat]
                vard=ror.variables[var]['data']
                varu=ror.variables[var]['units']

                if quiver:
                    if "_x" in var or "_y" in var or"_z" in var:
                        var2=var.split('_')[0]
                        qxvar=var2+"_x"
                        qyvar=var2+"_y"
                        qzvar=var2+"_z"
                        qxvard=ror.variables[qxvar]['data']
                        qyvard=ror.variables[qyvar]['data']
                        qzvard=ror.variables[qzvar]['data']
                        vard = np.sqrt(np.square(qxvard) +\
                                       np.square(qyvard) +\
                                       np.square(qzvard))

                if ror.coords[coord]['size'] == 1:
                    x=ror.variables[ror.coords[coord]['x']]['data']
                    y=ror.variables[ror.coords[coord]['y']]['data']
                    z=ror.variables[ror.coords[coord]['z']]['data']
                elif ror.coords[coord]['size'] == 3:
                    x=ror.variables[ror.coords[coord]['x']]['data'][:,0]
                    y=ror.variables[ror.coords[coord]['y']]['data'][:,1]
                    z=ror.variables[ror.coords[coord]['z']]['data'][:,2]
                bodyscale=1.
                if scale == "km":
                    if ror.variables[ror.coords[coord]['x']]['units'] == "R_E":
                        x=x*ror.RE
                        y=y*ror.RE
                        z=z*ror.RE
                        bodyscale=ror.RE    
                    r=(np.sqrt(x**2 + y**2 + z**2))-ror.RE
                else:
                    if ror.variables[ror.coords[coord]['x']]['units'] == "km":
                        x=x/ror.RE
                        y=y/ror.RE
                        z=z/ror.RE
                    r=(np.sqrt(x**2 + y**2 + z**2))-1.
            
                # Array of 'groupby' values and unique values
                if groupby == "day":
                    datearray = np.array([datetime.datetime.fromtimestamp(d,tz=timezone.utc).strftime\
                                          ("%Y-%m-%d") for d in ror.tsarray])
                elif groupby == "hour":
                    datearray = np.array([datetime.datetime.fromtimestamp(d,tz=timezone.utc).strftime\
                                          ("%Y-%m-%d %H") for d in ror.tsarray])
                else:
                    datearray = np.array([datetime.datetime.fromtimestamp(d,tz=timezone.utc).strftime\
                                          ("all") for d in ror.tsarray])
                udates = np.unique(datearray)
                nsteps = len(udates)

                for date in udates:
                    # Compute mast to restrict all data in trace
                    mask = date == datearray
                    # Create figure with data
                    if quiver:
                        fig.add_trace(go.Scatter3d(
                            name=date,  x=x[mask], y=y[mask], z=z[mask],
                            mode='markers',
                            marker=dict(
                                size=4,
                                cmin=cmin, cmax=cmax,
                                color=vard[mask],
                                showscale=True,
                                colorscale='Viridis',
                                colorbar=dict(title=var2+" ["+varu+"]"),
                            ),
                            customdata=np.vstack((ror.dtarrayclean[mask],\
                                                  qxvard[mask],\
                                                  qyvard[mask],\
                                                  qzvard[mask])).T,
                            hovertemplate="<b>"+ror.satellite+" Position</b>"+
                            "<br>X: %{x:.4f}<br>Y: %{y:.4f}<br>Z: %{z:.4f}<br>"+
                            qxvar+": %{customdata[1]:.2f}<br>"+
                            qyvar+": %{customdata[2]:.2f}<br>"+
                            qzvar+": %{customdata[3]:.2f}<br>"+
                            " "+var2+": %{marker.color:.2f}"+varu+"<br>"+
                            "%{customdata[0]}<br>"+"<extra></extra>",
                            visible=False,
                        ))
                        Nplot += 1
                    else:
                        fig.add_trace(go.Scatter3d(
                            name=date,  x=x[mask], y=y[mask], z=z[mask],
                            mode='markers',
                            marker=dict(
                                size=4,
                                cmin=cmin, cmax=cmax,
                                color=vard[mask],
                                showscale=True,
                                colorscale='Viridis',
                                colorbar=dict(title=var+" ["+varu+"]"),
                            ),
                            customdata=ror.dtarrayclean[mask],
                            hovertemplate="<b>"+ror.satellite+" Position</b>"+
                            "<br>X: %{x:.4f}<br>Y: %{y:.4f}<br>Z: %{z:.4f}<br>"+
                            " "+var+": %{marker.color:.2f}"+varu+"<br>"+
                            "%{customdata}<br>"+"<extra></extra>",
                            visible=False,
                        ))
                        Nplot += 1

                if quiver:
                    for date in udates:
                        # Compute mask to restrict all data in trace
                        mask = date == datearray
                        # Precompute needed values
                        xm=x[mask]
                        ym=y[mask]
                        zm=z[mask]
                        vm=vard[mask]
                        qxm=qxvard[mask]
                        qym=qyvard[mask]
                        qzm=qzvard[mask]
                        xm=xm.reshape(len(xm),1)
                        ym=ym.reshape(len(ym),1)
                        zm=zm.reshape(len(zm),1)
                        vm=vm.reshape(len(vm),1)
                        qxm=qxm.reshape(len(qxm),1)
                        qym=qym.reshape(len(qym),1)
                        qzm=qzm.reshape(len(qzm),1)
                        if quiverskip > 0:
                            for i in range(len(qxm)):
                                if i%(quiverskip+1) > 0:
                                    qxm[i]=0.
                                    qym[i]=0.
                                    qzm[i]=0.
                        xx=np.concatenate((xm,xm+qxm*quiverscale/cmax,xm),axis=1).reshape(3*len(xm))
                        yy=np.concatenate((ym,ym+qym*quiverscale/cmax,ym),axis=1).reshape(3*len(ym))
                        zz=np.concatenate((zm,zm+qzm*quiverscale/cmax,zm),axis=1).reshape(3*len(zm))
                        # Create figure with data
                        fig.add_trace(go.Scatter3d(
                            name=date,
                            x = xx,
                            y = yy,
                            z = zz,
                            mode='lines',
                            line=dict(
                                width=2,
                                color='rgba(22,22,22,0.2)',
                            ),
                            hoverinfo='skip',
                            visible=False,
                        ))
                        Nplot += 1

            # find cmin,cmax for traces and set to global min,max
            for i in range(len(fig.data)):
                if fig.data[i].marker.cmin is not None:
                    cmin=min(cmin,float(fig.data[i].marker.cmin))
                    cmax=max(cmax,float(fig.data[i].marker.cmax))
            for i in range(len(fig.data)):
                if fig.data[i].marker.cmin is not None:
                    fig.data[i].marker.cmin=cmin
                    fig.data[i].marker.cmax=cmax

            # find min,max locations to pad sphere points
            for i in range(len(fig.data)):
                xmin=min(xmin,np.amin(fig.data[i].x))
                xmax=max(xmax,np.amax(fig.data[i].x))
                ymin=min(ymin,np.amin(fig.data[i].y))
                ymax=max(ymax,np.amax(fig.data[i].y))
                zmin=min(zmin,np.amin(fig.data[i].z))
                zmax=max(zmax,np.amax(fig.data[i].z))
                
            # Add 1 RE sphere, padded to cover all data positions
            dataXYZ = pd.read_csv('https://ccmc.gsfc.nasa.gov/Kamodo/demo/sphereXYZ.csv')
            dataIJK = pd.read_csv('https://ccmc.gsfc.nasa.gov/Kamodo/demo/sphereIJK.csv')
            if scale == "km":
                dataXYZ=dataXYZ*self.RE
            fig.add_mesh3d()
            fig.data[Nplot].x = np.append(dataXYZ['x'],(xmin,xmax))
            fig.data[Nplot].y = np.append(dataXYZ['y'],(ymin,ymax))
            fig.data[Nplot].z = np.append(dataXYZ['z'],(zmin,zmax))
            fig.data[Nplot].i = dataIJK['i']
            fig.data[Nplot].j = dataIJK['j']
            fig.data[Nplot].k = dataIJK['k']
            fig.data[Nplot].facecolor = dataIJK['c']
            fig.data[Nplot].flatshading = True
            fig.data[Nplot].name = '1 R_E sphere'
            fig.data[Nplot].hovertemplate="Earth<extra></extra>"
            fig.data[Nplot].visible=True
            Nplot += 1

            # Start selection slider build
            steps = []
            for i in range(nsteps):
                step = dict(
                    method="update",
                    args=[{"visible": [False] * len(fig.data)}],
                    label=udates[i]
                )
                step["args"][0]["visible"][len(fig.data)-1] = True  # Set last trace to "visible"
                for j in range(len(self.satellites)):
                    if quiver:
                        step["args"][0]["visible"][i+(2*j*nsteps)] = True  # Toggle i'th trace
                        step["args"][0]["visible"][i+((2*j+1)*nsteps)] = True  # Toggle quiver trace
                    else:
                        step["args"][0]["visible"][i+(j*nsteps)] = True  # Toggle i'th trace
                steps.append(step)

            # Repeat first iteration of loop to set who is visible from start
            i=0
            fig.data[len(fig.data)-1].visible=True  # Set last trace to "visible"
            for j in range(len(self.satellites)):
                if quiver:
                    fig.data[i+(2*j*nsteps)].visible = True  # Toggle i'th trace
                    fig.data[i+((2*j+1)*nsteps)].visible = True  # Toggle quiver trace
                else:
                    fig.data[i+(j*nsteps)].visible = True  # Toggle i'th trace
            
            sliders = [dict(
                active=0,
                currentvalue={"prefix": "Currently showing: "},
                pad={"t": 50},
                steps=steps
            )]
            if nsteps > 1:
                fig.update_layout(sliders=sliders)

            # Update axis labels and plot title
            fig.update_layout(scene_aspectmode='data')
            fig.update_layout(scene=dict(xaxis=dict(title=dict(text="X ["+scale+"]")),
                                         yaxis=dict(title=dict(text="Y ["+scale+"]")),
                                         zaxis=dict(title=dict(text="Z ["+scale+"]"))))
            fig.update_layout(title_text=txttop)
            fig.update_layout(showlegend=False)
            fig.update_layout(scene_camera=dict(center=dict(x=0, y=0, z=0)))

            return fig



# -*- coding: utf-8 -*-
"""
Created on Fri Jan  6 10:21:46 2023

@author: rringuet

"""
from os.path import basename, isfile
from numpy import array, zeros, float32, mean
import h5py
from netCDF4 import Dataset


def GridRanges(data_files):
    '''Finds the max and min of each coordinate grid across the entire dataset.
    Writes the result to a simple text file named 'GAMERA_GridRanges.txt' in
    the same directory as the dataset. Called by the block of logic in the
    gamera reader that creates the time files.

    Input (data_files) - a list of the data files in the dataset
    Output: None, but writes the text file described.'''

    # calculate the max and min of each coordinate across the dataset
    x_max, x_min, y_max, y_min, z_max, z_min = [], [], [], [], [], []
    coord_shape = []
    for f in data_files:
        h5_data = h5py.File(f)
        X = array(h5_data['X'])
        Y = array(h5_data['Y'])
        Z = array(h5_data['Z'])
        h5_data.close()
        x_min.append(X.min())
        x_max.append(X.max())
        y_min.append(Y.min())
        y_max.append(Y.max())
        z_min.append(Z.min())
        z_max.append(Z.max())
        coord_shape.append(list(X.shape))
    del X, Y, Z
    X_min, X_max = array(x_min).min(), array(x_max).max()
    Y_min, Y_max = array(y_min).min(), array(y_max).max()
    Z_min, Z_max = array(z_min).min(), array(z_max).max()

    # write results to a simple text file
    file_dir = data_files[0].split(basename(data_files[0]))[0]
    files_str = ''.join([basename(f)+',' for f in data_files])[:-1]
    out_file = file_dir + 'GAMERA_GridRanges.txt'
    out = open(out_file, 'w')
    out.write(f'GAMERA_GridRanges output for files: {files_str}\n')
    out.write(f'X_min = {X_min:.5f}\n')
    out.write(f'X_max = {X_max:.5f}\n')
    out.write(f'Y_min = {Y_min:.5f}\n')
    out.write(f'Y_max = {Y_max:.5f}\n')
    out.write(f'Z_min = {Z_min:.5f}\n')
    out.write(f'Z_max = {Z_max:.5f}\n')
    out.close()

    return None


def Read_GridRanges(file_dir):
    '''Reads in the previously calculated grid ranges from the file produced by
    the GridRanges function above.
    Input: file_dir (the complete file path where the data is stored)
    Outputs:
        X_min - float value indicating the minimum X coordinate value across
            the dataset in the given file directory.
        X_max - float value indicating the maximum X coordinate value across
            the dataset in the given file directory.
        Y_min - float value indicating the minimum Y coordinate value across
            the dataset in the given file directory.
        Y_max - float value indicating the maximum Y coordinate value across
            the dataset in the given file directory.
        Z_min - float value indicating the minimum Z coordinate value across
            the dataset in the given file directory.
        Z_max - float value indicating the maximum Z coordinate value across
            the dataset in the given file directory.
    '''

    range_file = file_dir + 'GAMERA_GridRanges.txt'
    range_data = open(range_file, 'r')
    for line in range_data.readlines():
        if line[0] == 'G':
            continue
        tmp = float(line.split('=')[1].strip())
        if line[:5] == 'X_min':
            X_min = tmp
        if line[:5] == 'X_max':
            X_max = tmp
        if line[:5] == 'Y_min':
            Y_min = tmp
        if line[:5] == 'Y_max':
            Y_max = tmp
        if line[:5] == 'Z_min':
            Z_min = tmp
        if line[:5] == 'Z_max':
            Z_max = tmp
    range_data.close()

    return X_min, X_max, Y_min, Y_max, Z_min, Z_max


# Is this even necessary? Does the interpolator require the entire dataset
# for a variable to be assembled, or does it load sections lazily?
# Eric Winter: The interpolator loads the sections as needed.

# This calculates the cell centers and loads the coordinate grids.
# Need the entire coordinate grid to understand where the data pieces are.
# The requested variables need to have the same grid size as each other.
# Cannot ask for both X and dV because their sizes are different by 1 in each
# block.
def GridMapping(data_files, constant, sample_var):
    '''Determine what grid sections are in what files. Returns index mapping
    dictionary of beginning and ending indices for each block. Note that the
    constant variables are one size smaller than the coordinates because they
    are on cell-centers while the coordinates are the cell edges. Otherwise,
    the variables all have the same size, so the same index mapping may be used
    for all of them. This is called to map the coordinate variables together
    for calculation of the cell centers in the pre-processing step. Avoids
    converting data to arrays, so execution time it quite fast.
    Inputs:
        data_files - a list of the h5 files containing the data
        constant - boolean, whether the variable is constant or not. This
            decides whether the program looks in the 'Step#0' group or the main
            group for the data.
        sample_var - the name of a variable with the same grid (string
    Example usage of returned dictionary:
        net_array = np.zeros(net_indices['all'][1])
        net_array[net_indices[key][0][0]:net_indices[key][1][0]] = data_block
        for a 1D array, where data_block is from the file corresponding to the
        key name, and key is the block number of format '00xi_00yi_00zi'.
    '''
    # first, get file grid from filename of first file
    # pattern: abcd_00nx_00ny_00nz_0nxi_0nyi_0nzi....
    nx, ny, nz = array(basename(data_files[0]).split('_')[1:4],
                       dtype=int)
    # loop through files and get size of each grid
    indices = {}
    for f in data_files:
        key = ''.join([b+'_' for b in basename(f).split('.')[0].split('_')[-3:]
                       ])[:-1]  # 0000_0000_0000 for first block, etc
        h5_data = h5py.File(f)
        if constant:
            indices[key] = h5_data[sample_var].shape
        else:
            indices[key] = h5_data['Step#0'][sample_var].shape
        h5_data.close()

    # then get the total shape by summing the data shapes
    # this assumes that the assembled block will have a single (x, y, z) shape
    x_shape = sum([indices[str(i).zfill(4)+'_0000_0000'][0] for i in
                   range(nx)])
    y_shape = sum([indices['0000_'+str(i).zfill(4)+'_0000'][1] for i in
                   range(ny)])
    z_shape = sum([indices['0000_0000_'+str(i).zfill(4)][2] for i in
                   range(nz)])

    # determine beginning and ending indices for each block
    net_indices = {'net': [[0, 0, 0], [x_shape, y_shape, z_shape]]}
    for key in indices.keys():
        # get beginning position mapping in array
        nxi, nyi, nzi = array(key.split('_'), dtype=int)
        xi = sum([indices[str(i).zfill(4)+'_0000_0000'][0] for i in
                  range(nxi)])
        yi = sum([indices['0000_'+str(i).zfill(4)+'_0000'][1] for i in
                  range(nyi)])
        zi = sum([indices['0000_0000_'+str(i).zfill(4)][2] for i in
                  range(nzi)])
        # get ending position+1
        xip1, yip1, zip1 = array([xi, yi, zi]) + indices[key]
        net_indices[key] = [[xi, yi, zi], [xip1, yip1, zip1]]
    print(net_indices)
    return net_indices


def AssembleGrid(data_files, net_indices, constant, var, step=0):
    '''Assemble the requested variable into a single grid from all the files.
    Used for the coordinate assembly. This assumes that there are no ghost
    cells, which is often an incorrect assumption.

    Inputs:
        data_files - a list of h5 files containing the data to be assembled
        net_indices - the output from the GridMapping function above
        constant - boolean, whether the variable is constant or not. This
            decides whether the program looks in the 'Step#?' group or the main
            group for the data.
        var - a string indicating the name of the desired variable in the files
        step - an integer indicating the time step of the data to be assembled,
            default is zero.
    Output:
        A float32 array of the assembled data.
    '''

    data = zeros(net_indices['net'][1], dtype=float32)
    for f in data_files:
        key = ''.join([b+'_' for b in basename(f).split('.')[0].split('_')[-3:]
                       ])[:-1]  # 0000_0000_0000 for first block, etc
        h5_data = h5py.File(f)
        x0, y0, z0 = net_indices[key][0]
        x1, y1, z1 = net_indices[key][1]
        if constant:
            data[x0:x1, y0:y1, z0:z1] = array(h5_data[var])
        else:
            data[x0:x1, y0:y1, z0:z1] = array(h5_data['Step#'+str(step)][var])
        h5_data.close()
    return data


def ComputeCellCentersFile(data_files):
    '''Computes the centers of the X, Y, and Z coordinate grids individually
    for each file. Avoids the complication of ghost cells during grid assembly.
    Input:
        data_files - list of h5 files containing the data
    Outputs: None, but writes a set of files ending in 'gridcenters.nc'
        containing the following variables for each file.
        X_c - numpy array of X locations of cell centers, one size smaller than
            the assembled X grid from the file(s)
        Y_c - numpy array of Y locations of cell centers, one size smaller than
            the assembled Y grid from the file(s)
        Z_c - numpy array of Z locations of cell centers, one size smaller than
            the assembled Z grid from the file(s)
    '''

    # read in data from each file
    file_dir = data_files[0].split(basename(data_files[0]))[0]
    for f in data_files:
        center_file = file_dir + basename(f).split('.')[0] + '_gridcenters.nc'
        if isfile(center_file):
            continue
        h5_data = h5py.File(f)
        X = array(h5_data['X'])
        Y = array(h5_data['Y'])
        Z = array(h5_data['Z'])
        h5_data.close()

        # compute cell centers
        X_c = zeros([X.shape[0]-1, X.shape[1]-1, X.shape[2]-1])
        Y_c, Z_c = X_c.copy(), X_c.copy()
        for i in range(X_c.shape[0]):
            for j in range(X_c.shape[1]):
                for k in range(X_c.shape[2]):
                    X_c[i, j, k] = mean(X[i:i+2, j:j+2, k:k+2])
                    Y_c[i, j, k] = mean(Y[i:i+2, j:j+2, k:k+2])
                    Z_c[i, j, k] = mean(Z[i:i+2, j:j+2, k:k+2])

        # prepare output file
        data_out = Dataset(center_file, 'w', format='NETCDF4')
        data_out.Description = 'Cell centers of the coordinate grids for ' + f

        # create dimensions, all coord grids have the same shapes,
        # but shapes could be diff in diff files
        new_dim = data_out.createDimension('c1', X_c.shape[0])
        new_dim = data_out.createDimension('c2', X_c.shape[1])
        new_dim = data_out.createDimension('c3', X_c.shape[2])
        # save X center positions
        new_var = data_out.createVariable('X_c', float32, ('c1', 'c2', 'c3'))
        new_var[:] = X_c
        new_var.units = 'R_E'
        new_var.Description = 'X coordinate of grid cell centers'
        # save Y center positions
        new_var = data_out.createVariable('Y_c', float32, ('c1', 'c2', 'c3'))
        new_var[:] = Y_c
        new_var.units = 'R_E'
        new_var.Description = 'Y coordinate of grid cell centers'
        # save Z center positions
        new_var = data_out.createVariable('Z_c', float32, ('c1', 'c2', 'c3'))
        new_var[:] = Z_c
        new_var.units = 'R_E'
        new_var.Description = 'Z coordinate of grid cell centers'
        # close file
        data_out.close()

    return None


def ComputeCellCentersTotal(data_files, write_file=True, grid=True):
    '''Computes the centers of the X, Y, and Z coordinate grids.
    Input:
        data_files - list of h5 files containing the data
        write_file - boolean (default=True), writes cells centers to a netCDF4
            file if True, doesn't if False.
    Outputs:
        X_c - numpy array of X locations of cell centers, one size smaller than
            the assembled X grid from the file(s)
        Y_c - numpy array of Y locations of cell centers, one size smaller than
            the assembled Y grid from the file(s)
        Z_c - numpy array of Z locations of cell centers, one size smaller than
            the assembled Z grid from the file(s)
    '''
    # assemble coordinate grid from the file(s)
    if len(data_files) > 1 and not grid:  # only do this if grid needs assembln
        # get index mapping for X, identical to that for Y and Z
        net_indices = GridMapping(data_files, True, 'X')
        # assemble coordinate grids from MPI files
        X = AssembleGrid(data_files, net_indices, True, 'X')
        Y = AssembleGrid(data_files, net_indices, True, 'Y')
        Z = AssembleGrid(data_files, net_indices, True, 'Z')
    else:
        # read in data from serial file
        h5_data = h5py.File(data_files[0])
        X = array(h5_data['X'])
        Y = array(h5_data['Y'])
        Z = array(h5_data['Z'])
        h5_data.close()

    # compute cell centers
    X_c = zeros([X.shape[0]-1, X.shape[1]-1, X.shape[2]-1])
    Y_c, Z_c = X_c.copy(), X_c.copy()
    for i in range(X_c.shape[0]):
        for j in range(X_c.shape[1]):
            for k in range(X_c.shape[2]):
                X_c[i, j, k] = mean(X[i:i+2, j:j+2, k:k+2])
                Y_c[i, j, k] = mean(Y[i:i+2, j:j+2, k:k+2])
                Z_c[i, j, k] = mean(Z[i:i+2, j:j+2, k:k+2])

    # write to file if requested
    if write_file:
        from netCDF4 import Dataset

        file_dir = data_files[0].split(basename(data_files[0]))[0]
        center_file = file_dir + 'gridcenters.nc'
        data_out = Dataset(center_file, 'w', format='NETCDF4')
        data_out.Description = 'Cell centers of the coordinate grids.'
        # create dimensions, all coord grids have the same shapes
        new_dim = data_out.createDimension('c1', X_c.shape[0])
        new_dim = data_out.createDimension('c2', X_c.shape[1])
        new_dim = data_out.createDimension('c3', X_c.shape[2])
        # save X center positions
        new_var = data_out.createVariable('X_center', float32,
                                          ('c1', 'c2', 'c3'))
        new_var[:] = X_c
        new_var.units = 'R_E'
        new_var.Description = 'X coordinate of grid cell centers'
        # save Y center positions
        new_var = data_out.createVariable('Y_center', float32,
                                          ('c1', 'c2', 'c3'))
        new_var[:] = Y_c
        new_var.units = 'R_E'
        new_var.Description = 'Y coordinate of grid cell centers'
        # save Z center positions
        new_var = data_out.createVariable('Z_center', float32,
                                          ('c1', 'c2', 'c3'))
        new_var[:] = Z_c
        new_var.units = 'R_E'
        new_var.Description = 'Z coordinate of grid cell centers'
        # close file
        data_out.close()

    return X_c, Y_c, Z_c



# -*- coding: utf-8 -*-
"""
Created on Tue Sep 20 11:57:13 2022

@author: rringuet
"""
from csv import reader
from time import perf_counter
from datetime import datetime, timezone
from netCDF4 import Dataset
from numpy import array, float32, unique, append, reshape
from os.path import getsize
from statistics import mode
from kamodo_ccmc.readers.reader_utilities import str_to_hrs


def convert_all(files):
    '''Converting all files into one netCDF4. Skips files of different sizes
    to avoid file errors. Typically, this is the first file and any list files.
    '''

    sizes = [getsize(f)/1024. for f in files]  # file sizes in KB
    good_files = [f for i, f in enumerate(files) if sizes[i] == mode(sizes)]
    bad_files = [f for f in files if f not in good_files]
    if len(bad_files) > 0:
        print('Skipping the following files: ', bad_files)
    cdf_file = good_files[0][:-18] + '.nc'
    to_CDF(cdf_file, good_files)  # only perform analysis on good data
    return True


def read_weimerfile(file_name):
    '''Read in data from given file and return in a dictionary of arrays.
    Weimer data starts at 12 MLT and ends at 12 MLT. Longitude is already
    wrapped in the data.'''

    # read in data from file
    file_obj = open(file_name, 'r')
    csv_reader = reader(file_obj, delimiter='\n')  # splits file by \n char

    # skip header
    for i in range(9):
        line = next(csv_reader)
    # initialize based on what variables are present
    data = {key: [] for key in line[0][1:].split()}
    var_list = list(data.keys())
    line = next(csv_reader)  # skip units line
    # read in data
    for line in csv_reader:
        for i, val in enumerate(line[0].split()):
            data[var_list[i]].append(val)
    for key in data.keys():
        data[key] = array(data[key], dtype=float32)
    file_obj.close()
    r = data['R'][0]  # constant value
    del data['R']
    return data, r


def to_CDF(cdf_filename, files):
    '''Given the filename of the new cdf file and the list of files, put all
    the data into a single netcdf4 file.'''

    print('Converted data file not found. Converting files with ' +
          f'{cdf_filename.split(".")[0]} naming pattern to a netCDF4 file.')
    time0 = perf_counter()
    # choosing netCDF format that works for large files
    data_out = Dataset(cdf_filename, 'w', format='NETCDF3_64BIT_OFFSET')
    data_out.file = ''.join([f+',' for f in files])[:-1]
    data_out.model = 'Weimer'

    # figure out date and time information from filenames
    filedate = datetime.strptime(
        files[0][-17:-9]+' 00:00:00', '%Y%m%d %H:%M:%S').replace(
            tzinfo=timezone.utc)
    data_out.filedate = filedate.strftime('%Y-%m-%d %H:%M:%S')
    date_strs = [f[-17:-4] for f in files]
    time = str_to_hrs(date_strs, filedate, '%Y%m%d_%H%M')

    # establish time coordinate (time, lon, lat)
    # time: create dimension and variable
    new_dim = data_out.createDimension('time', len(files))
    new_var = data_out.createVariable('time', float32, 'time')
    new_var[:] = time
    new_var.units = 'hr'

    # get lon and lat information from first file
    data, r = read_weimerfile(files[0])
    var_list = [var for var in data.keys() if var not in ['LAT', 'MLT']]
    data_out.radius = float32(r)
    tmp = unique(data['MLT']) * 15.  # files starts at -180 and ends at 180
    lon = append(tmp - 180., 180.)  # lon already wrapped in file
    lat = unique(data['LAT'])

    # lon: create dimension and variable
    new_dim = data_out.createDimension('lon', lon.size)
    new_var = data_out.createVariable('lon', float32, 'lon')
    new_var[:] = lon  # store data for dimension in variable
    new_var.units = 'deg'
    # lat: create dimension and variable
    new_dim = data_out.createDimension('lat', lat.size)
    new_var = data_out.createVariable('lat', float32, 'lat')
    new_var[:] = lat  # store data for dimension in variable
    new_var.units = 'deg'

    # copy over variable to file (only first time in data['PHI'])
    for var in var_list:
        new_var = data_out.createVariable(var, float32,
                                          ('time', 'lon', 'lat'))
        new_var[0] = reshape(data[var], (lon.shape[0], lat.shape[0]))

    # loop through remaining files
    for i, f in enumerate(files[1:]):
        data, r = read_weimerfile(f)
        for var in var_list:
            data_out[var][i+1] = reshape(data[var],
                                         (lon.shape[0], lat.shape[0]))

    # close file
    data_out.close()
    print('Time to create netCDF4 file:', perf_counter()-time0)
    return None



# -*- coding: utf-8 -*-
"""
Created on Mon May 17 18:53:35 2021
@author: rringuet
"""
from glob import glob
import numpy as np
from time import perf_counter
from os.path import basename, isfile
from netCDF4 import Dataset

swmfie_varnames = {"X": ['x', 'km'], "Y": ['y', 'km'], "Z": ['z', 'km'],
                   # coordinates are ignored
                   "Theta": ['theta', "deg"], "Psi": ['psi', "deg"],
                   "Btilt_theta": ['theta_Btilt', "deg"],
                   "Btilt_psi": ['psi_Btilt', "deg"],
                   "SigmaH": ['Sigma_H', "S"], "SigmaP": ['Sigma_P', "S"],
                   "E-Flux": ['Phi_E', "W/m**2"], "Ave-E": ['E_avg', 'eV'],
                   "JR": ["j_R", "mA/m**2"], "PHI": ["Phi", "kV"],
                   "Ex": ["E_x", "mV/m"], "Ey": ["E_y", "mV/m"],
                   "Ez": ["E_z", "mV/m"],
                   "Jx": ["j_x", "mA/m**2"], "Jy": ["j_y", "mA/m**2"],
                   "Jz": ["j_z", "mA/m**2"],
                   "Ux": ['v_x', "km/s"], "Uy": ['v_y', "km/s"],
                   "Uz": ['v_z', "km/s"],
                   "JouleHeat": ['Q_Joule', "mW/m**2"],
                   "IonNumFlux": ['Phi_nion', "1/cm**2/s"],
                   "RT 1/B": ['Binv_RT', "1/T"],
                   "RT Rho": ['rho_RT', "amu/cm**3"],
                   "RT P": ['P_RT', "Pa"],
                   "conjugate dLat": ['dLat_star', "deg"],
                   "conjugate dLon": ['dlon_star', "deg"]}

'''
Documentation variables:
(ignored, given in R_E on a unit sphere)
"X":['x','km'],"Y":['y','km'],"Z":['z','km'],
"Theta":['theta',"deg"],"Psi":['psi',"deg"],     (used as coordinates)
"Btilt_theta":['theta_Btilt',"deg"], "Btilt_psi":['psi_Btilt',"deg"]
(added directly to object for documentation purposes)
'''


def convert_all(file_dir):
    '''Convert all files independent of the others.'''

    print('NetCDF version of data not found. Converting tec files in ' +
          f'{file_dir} to netCDF.')
    ftic, nfiles = perf_counter(), 0
    files = sorted(glob(file_dir + '*.tec'))  # want YYYYMMDD
    if len(files) == 0:
        print('No original files found.')
        return

    # convert and return
    for file in files:
        if not isfile(file.replace('.tec', '.nc')):
            coords, variables, var_units, theta_Btilt, psi_Btilt = \
                _read_SWMFIE(file, verbose=True)
            cdf_filename = _toCDF(file, coords, variables, var_units,
                                  theta_Btilt, psi_Btilt)
            nfiles += 1
            print(cdf_filename, ' converted.')
    print(f'{nfiles} files in {file_dir} now converted into ' +
          f'netCDF4 files in {perf_counter()-ftic:.6f}s.')
    return None


def read_SWMFIE_header(filename):
    '''parse header for a representative file'''

    # get file data
    file_object = open(filename, 'r')
    file_contents = file_object.read()
    file_object.close()

    # sort info with TITLE keyword
    title_lines = file_contents.split('TITLE=')[1].split('VARIABLES=')[0].\
        replace('\n', ',').replace('"', '').strip().strip(',')
    title, time, BtiltDeg_str = title_lines.split(',')
    # get time into 'YYYY-MM-DD HH:mm:SS' format, clean up BtiltDeg
    dts = time[:10]+' '+''.join([time[11:].split('-')[i]+':'
                                 for i in [0, 1]])+time[11:].split('-')[2]
    BtiltDeg = BtiltDeg_str.split('=')[1].split()

    # sort info with VARIABLE keyword
    variable_lines = file_contents.split('VARIABLES=')[1].split('ZONE')[0].\
        replace('\n', ',').replace('"', '').strip().strip(',')
    variable_list = [var.strip().split(']')[0].split(' [')[0] for var in
                     variable_lines.split(',')]  # units in dict above

    # sort info with ZONE keyword 'IonN N=0000241 T=0000:20:00'
    zone_lines = file_contents.split('ZONE T=')[1].split('\n')[0].strip(
        '"').strip()
    run_type1, N1, t = zone_lines.split(' ')
    N1 = int(N1.split('=')[1].strip())
    zone_lines = file_contents.split('ZONE T=')[2].split('\n')[0].strip(
        '"').strip()
    run_type2, N2, t = zone_lines.split(' ')
    N2 = int(N2.split('=')[1].strip())

    # sort coordinate info   'I=           91  J=          181  F=POINT'
    coord_info = file_contents.split(zone_lines)[1].split('\n')[1].strip()
    text, i, text, j, f = coord_info.split()
    i, j, f = int(i), int(j), f.split('=')[1]
    # if sometimes different, need another set for this

    # determine number of lines to skip for data sections
    line_test = [f in c for c in file_contents.split('\n')]
    skip1, skip2 = np.where(np.array(line_test, dtype=int) == 1)[0]

    # determine how many lines constitute one data list
    data_lines = file_contents.split('\n')[skip1+1:]
    num_test = [len(line) for line in data_lines]  # list of lengths of lines
    ndata_lines = np.diff(np.where(np.array(num_test) == num_test[0])[0])[0]

    # cleanup and return
    del file_contents
    header = {'title': title, 'time': dts, 'BtiltDeg': BtiltDeg,
              'variable_list': variable_list, 'run_type1': run_type1, 'N1': N1,
              'run_type2': run_type2, 'N2': N2, 'i': i, 'j': j, 'f': f,
              'skip1': skip1, 'skip2': skip2, 'ndata_lines': ndata_lines}
    return header


def read_SWMFIE_data(filename, header):
    '''read only data from file and return in labeled dictionary'''

    # get data from file using skip as number of lines to skip
    file_object = open(filename, 'r')
    file_contents = file_object.readlines()
    file_object.close()

    # Get theta_Btilt value from file
    title_line = file_contents[0].replace('\n',
                                          ',').replace('"',
                                                       '').strip().strip(',')
    theta_Btilt = float(title_line.split(',')[-1].strip().split()[1])

    # collect data into an array, one for each run_type
    file_data1 = file_contents[header['skip1']+1:header['skip2']-1]
    data1 = np.array([''.join(file_data1[i:i+header['ndata_lines']]).
                      strip('\n').split() for i in range(
                          0, len(file_data1), header['ndata_lines'])],
                     dtype=float)
    file_data2 = file_contents[header['skip2']+1:]
    data2 = np.array([''.join(file_data2[i:i+header['ndata_lines']]).
                      strip('\n').split() for i in range(
                          0, len(file_data2), header['ndata_lines'])],
                     dtype=float)

    # determine correct dimension sizes from theta and psi (3 and 4)
    # combine data and return in dict
    nLatA, nLon = np.unique(data1[:, 3]).size, np.unique(data1[:, 4]).size
    data = np.concatenate((np.reshape(data1, (nLon, nLatA,
                                              len(header['variable_list']))),
                           np.reshape(data2, (nLon, nLatA,
                                              len(header['variable_list'])))),
                          axis=1)[:, ::-1, :]
    # leave since grids are identical in N/S hemispheres

    return {swmfie_varnames[header['variable_list'][i]][0]:
            np.transpose(data[:, :, i], [1, 0])
            for i in range(len(header['variable_list']))}, theta_Btilt


def _read_SWMFIE(file, verbose=False):
    '''file_dir should be the complete path to the location of the files.'''
    t0 = perf_counter()

    # establish time attributes first for file searching
    file_datestr = basename(file)[3:11]
    # string_date = 'YYYY-MM-DD'
    string_date = file_datestr[:4]+'-'+file_datestr[4:6]+'-'+file_datestr[6:8]

    # get list of variables possible in these files using first file
    file_header = read_SWMFIE_header(file)
    file_varlist = file_header['variable_list']
    gvar_list = [swmfie_varnames[key][0] for key in file_varlist if key not in
                 ['X', 'Y', 'Z', 'Theta', 'Psi', 'Btilt_theta', 'Btilt_psi']]
    # avoid returning coordinates stored elsewhere (or ignored)

    # collect coordinates and intialize data storage from first file
    data, Btilt = read_SWMFIE_data(file, file_header)
    lon, lat0 = np.unique(data['psi']), np.unique(data['theta'])-90.
    lon_ge180 = list(np.where((lon >= 180.) & (lon < 360.))[0])
    lon_le180 = list(np.where(lon <= 180.)[0])
    lon_idx = lon_ge180 + lon_le180
    lon -= 180  # shifting longitude to be in range -180 to 180
    lat0 = np.insert(lat0, np.where(lat0 == 0.)[0], 0.)
    # interpolator requires unique ascending values
    # offset two zero values by 0.0001
    lat0[np.where(lat0 == 0.)[0]] = -0.0000001, 0.0000001
    lat = np.insert(lat0, 0, -90.)
    lat = np.append(lat, 90.)  # add values for poles for wrapping
    # data is 2D (lon, lat) with time, so no height
    theta_Btilt = Btilt   # to append more values later
    psi_Btilt = 0.   # this value is always zero

    # Store variable data and units for the file.
    # transpose (lat, lon) -> (lon, lat) and wrap longitude
    var_units = {value[0]: value[-1] for key, value in swmfie_varnames.items()
                 if value[0] in gvar_list}
    variables = {key: np.array(data[key]).T[lon_idx] for key in gvar_list}

    # wrap latitude for scalar variables (all to be treated as scalars)
    for var_key in variables:
        tmp = variables[var_key]
        new_shape = list(tmp.shape)
        new_shape[1] += 2  # one value on each end
        tmp2 = np.zeros(new_shape)
        tmp2[:, 1:-1] = tmp
        # put in top values
        top = np.mean(tmp2[:, 1], axis=0)  # single value
        new_top = np.broadcast_to(top, (new_shape[0]))
        tmp2[:, 0] = new_top
        # same for bottom, reusing variable names
        top = np.mean(tmp2[:, -2], axis=0)  # single value
        new_top = np.broadcast_to(top, (new_shape[0]))
        tmp2[:, -1] = new_top
        variables[var_key] = tmp2  # store result

    coords = {'lat': lat, 'lon': lon}
    return coords, variables, var_units, theta_Btilt, psi_Btilt
# removed files from return statement


def _toCDF(file, coords, variables, var_units, theta_Btilt, psi_Btilt):
    '''Write data to a netCDF4 file.'''

    # start new wrapped output object
    cdf_filename = file.replace('.tec', '.nc')
    data_out = Dataset(cdf_filename, 'w', format='NETCDF4')
    data_out.file = file
    data_out.model = 'SWMF_IE'
    for dim in coords.keys():
        units = 'deg'
        # create dimension
        new_dim = data_out.createDimension(dim, coords[dim].size)
        # create variable
        new_var = data_out.createVariable(dim, np.float64, dim)
        new_var[:] = coords[dim]  # store data for dimension in variable
        new_var.units = units

    # copy over variables to file
    for variable_name in variables.keys():
        if len(variables[variable_name].shape) == 2:
            new_var = data_out.createVariable(variable_name, np.float64,
                                              ('lon', 'lat'))
            new_data = variables[variable_name]
        else:
            continue
        new_var[:] = new_data  # store data in variable
        new_var.units = var_units[variable_name]
    data_out.close()
    return cdf_filename



from datetime import datetime, timezone


# varnames in cdf files are standardized (value[0])
model_varnames = {'den400': ['rho_400km', 'Density at 400 km.',
                             0, 'GDZ', 'sph', ['time', 'lon', 'lat'],
                             'kg/m**3'],
                  'ON2': ['ON2', 'mmr or ratio?',
                          0, 'GDZ', 'sph', ['time', 'lon', 'lat'], 'm/m'],
                  'height': ['H_ilev', 'Height dependent upon pressure level',
                             0, 'GDZ', 'sph', ['time', 'lon', 'lat', 'ilev'],
                             'm'],
                  'temp_neutral': ['T_n_ilev', 'neutral temperature', 0, 'GDZ',
                                   'sph', ['time', 'lon', 'lat', 'ilev'],
                                   'K'],
                  'temp_neutral_2': ['T_n', 'neutral temperature', 0, 'GDZ',
                                     'sph', ['time', 'lon', 'lat', 'height'],
                                     'K'],
                  'O_Density': ['N_O_ilev', 'Oxygen number density', 0, 'GDZ',
                                'sph', ['time', 'lon', 'lat', 'ilev'],
                                '1/m**3'],
                  'O_Density_2': ['N_O', 'Oxygen number density', 0, 'GDZ',
                                  'sph', ['time', 'lon', 'lat', 'height'],
                                  '1/m**3'],
                  'O2_Density': ['N_O2_ilev', 'Molecular oxygen number ' +
                                 'density', 0, 'GDZ', 'sph',
                                 ['time', 'lon', 'lat', 'ilev'], '1/m**3'],
                  'O2_Density_2': ['N_O2', 'Molecular oxygen number density',
                                   0, 'GDZ', 'sph', ['time', 'lon', 'lat',
                                                     'height'], '1/m**3'],
                  'N2_Density': ['N_N2_ilev', 'Molecular nitrogen number ' +
                                 'density', 0, 'GDZ', 'sph',
                                 ['time', 'lon', 'lat', 'ilev'], '1/m**3'],
                  'N2_Density_2': ['N_N2', 'Molecular nitrogen number density',
                                   0, 'GDZ', 'sph', ['time', 'lon', 'lat',
                                                     'height'], '1/m**3'],
                  'u_neutral': ['v_neast_ilev', 'Eastward component of the ' +
                                'neutral wind velocity', 0, 'GDZ', 'sph',
                                ['time', 'lon', 'lat', 'ilev'], 'm/s'],
                  'u_neutral_2': ['v_neast', 'Eastward component of the ' +
                                  'neutral wind velocity', 0, 'GDZ', 'sph',
                                  ['time', 'lon', 'lat', 'height'], 'm/s'],
                  'v_neutral': ['v_nnorth_ilev', 'Northward component of the' +
                                ' neutral wind velocity', 0, 'GDZ', 'sph',
                                ['time', 'lon', 'lat', 'ilev'], 'm/s'],
                  'v_neutral_2': ['v_nnorth', 'Northward component of the ' +
                                  'neutral wind velocity', 0, 'GDZ', 'sph',
                                  ['time', 'lon', 'lat', 'height'], 'm/s'],
                  'w_neutral': ['v_nup_ilev', 'Upwards component of the ' +
                                'neutral wind velocity', 0, 'GDZ', 'sph',
                                ['time', 'lon', 'lat', 'ilev'], 'm/s'],
                  'w_neutral_2': ['v_nup', 'Upwards component of the ' +
                                  'neutral wind velocity', 0, 'GDZ', 'sph',
                                  ['time', 'lon', 'lat', 'height'], 'm/s'],
                  'tec': ['TEC', 'Total electron content', 0, 'GDZ', 'sph',
                          ['time', 'lon', 'lat'], '10**16/m**2'],
                  'NmF2': ['NmF2', 'Maximum number density in the F2 layer',
                           0, 'GDZ', 'sph', ['time', 'lon', 'lat'], '1/m**3'],
                  'HmF2': ['HmF2', 'Height of the maximum number density in ' +
                           'the F2 layer', 0, 'GDZ', 'sph',
                           ['time', 'lon', 'lat'], 'km'],
                  'O_plus_density': ['N_Oplus', 'O+ number density', 0, 'GDZ',
                                     'sph', ['time', 'lon', 'lat', 'height'],
                                     '1/m**3'],
                  'H_plus_density': ['N_Hplus', 'H+ number density', 0, 'GDZ',
                                     'sph', ['time', 'lon', 'lat', 'height'],
                                     '1/m**3'],
                  'He_plus_density': ['N_Heplus', 'He+ number density', 0,
                                      'GDZ', 'sph', ['time', 'lon', 'lat',
                                                     'height'], '1/m**3'],
                  'N_plus_density': ['N_Nplus', 'N+ number density', 0, 'GDZ',
                                     'sph', ['time', 'lon', 'lat', 'height'],
                                     '1/m**3'],
                  'NO_plus_density': ['N_NOplus', 'NO+ number density', 0,
                                      'GDZ', 'sph', ['time', 'lon', 'lat',
                                                     'height'], '1/m**3'],
                  'O2_plus_density': ['N_O2plus', 'O2+ number density', 0,
                                      'GDZ', 'sph', ['time', 'lon', 'lat',
                                                     'height'], '1/m**3'],
                  'N2_plus_density': ['N_O2plus', 'N2+ number density', 0,
                                      'GDZ', 'sph', ['time', 'lon', 'lat',
                                                     'height'], '1/m**3'],
                  'ion_temperature': ['T_i', 'Ion temperature', 0, 'GDZ',
                                      'sph', ['time', 'lon', 'lat', 'height'],
                                      'K'],
                  'electron_temperature': ['T_e', 'Electron temperature', 0,
                                           'GDZ', 'sph', ['time', 'lon', 'lat',
                                                          'height'], 'K'],
                  'eastward_exb_velocity': ['v_eastExB', 'Eastward component' +
                                            ' of the ExB velocity', 0, 'GDZ',
                                            'sph', ['time', 'lon', 'lat',
                                                    'height'], 'm/s'],
                  'northward_exb_velocity': ['v_northExB', 'Northward ' +
                                             'component of the ExB velocity',
                                             0, 'GDZ', 'sph',
                                             ['time', 'lon', 'lat', 'height'],
                                             'm/s'],
                  'upward_exb_velocity': ['v_upExB', 'Upward component' +
                                          ' of the ExB velocity', 0, 'GDZ',
                                          'sph', ['time', 'lon', 'lat',
                                                  'height'], 'm/s']
                  }


def MODEL():
    from time import perf_counter
    from glob import glob
    from os.path import basename, isfile
    from numpy import array, unique, NaN, append, linspace, where, median
    from numpy import zeros
    from netCDF4 import Dataset
    from kamodo import Kamodo
    import kamodo_ccmc.readers.reader_utilities as RU

    class MODEL(Kamodo):
        '''WAM-IPE model data reader.

        Inputs:
            file_dir: a string representing the file directory of the
                model output data.
                Note: This reader 'walks' the entire dataset in the directory.
            variables_requested = a list of variable name strings chosen from
                the model_varnames dictionary in this script, specifically the
                first item in the list associated with a given key.
                - If empty, the reader functionalizes all possible variables
                    (default)
                - If 'all', the reader returns the model_varnames dictionary
                    above for only the variables present in the given files.
            filetime = boolean (default = False)
                - If False, the script fully executes.
                - If True, the script only executes far enough to determine the
                    time values associated with the chosen data.
            printfiles = boolean (default = False)
                - If False, the filenames associated with the data retrieved
                    ARE NOT printed.
                - If True, the filenames associated with the data retrieved ARE
                    printed.
            gridded_int = boolean (default = True)
                - If True, the variables chosen are functionalized in both the
                    standard method and a gridded method.
                - If False, the variables chosen are functionalized in only the
                    standard method.
            verbose = boolean (False)
                - If False, script execution and the underlying Kamodo
                    execution is quiet except for specified messages.
                - If True, be prepared for a plethora of messages.
        All inputs are described in further detail in
            KamodoOnboardingInstructions.pdf.

        Returns: a kamodo object (see Kamodo core documentation) containing all
            requested variables in functionalized form.

        Notes:
            - WAM-IPE output files are given in multiple netCDF files per day.
              No attempt is made to combine these files, but some
              pre-processing is needed to calculate the median height across
              the entire dataset (time, lon, lat) for a given pressure level
              value. If a tarball of files is detected in the absense of the
              netCDF files, then the tarball will be decompressed and the
              netCDF files extracted automatically.
            - WAM-IPE data is given in several coordinate systems, one
              depending on pressure level - a unitless representation of height
              corresponding to a given atmospheric pressure. The pressure
              level values are not given in the data, so an arbitrary integer
              representation is chosen. 
            - Pressure level inversion is performed in the reader_utilities
              script, specifically the PLevelInterp function. Two versions of
              all variables that depend on pressure level are created: the
              original and one dependent on height, which is created through
              Kamodo's function composition feature.
            - The files are small and contain one time step per file, so
              interpolation method 1 is chosen. The standard SciPy interpolator
              is used.
        '''
        def __init__(self, file_dir, variables_requested=[],
                     filetime=False, verbose=False, gridded_int=True,
                     printfiles=False, **kwargs):
            super(MODEL, self).__init__()
            self.modelname = 'WAMIPE'
            t0 = perf_counter()

            # first, check for file list, create if DNE
            list_file = file_dir + self.modelname + '_list.txt'
            time_file = file_dir + self.modelname + '_times.txt'
            self.times, self.pattern_files = {}, {}
            if not isfile(list_file) or not isfile(time_file):
                # collect filenames
                files = sorted(glob(file_dir+'*.nc'))
                if len(files) == 0:  # find tar files and untar them
                    print('Decompressing files...this may take a moment.')
                    import tarfile
                    tar_files = glob(file_dir+'*.tar')
                    for file in tar_files:
                        tar = tarfile.open(file)
                        tar.extractall(file_dir)
                        tar.close()
                    files = sorted(glob(file_dir+'*.nc'))

                # create h0 file containing km_ilev
                from kamodo_ccmc.readers.wamipe_tocdf import convert_all
                tmp = convert_all(file_dir)

                # continue
                files = sorted(glob(file_dir+'*.nc'))
                h0_file = [f for f in files if 'h0' in f]
                if len(h0_file) > 0:
                    files.remove(h0_file[0])
                patterns = unique([basename(f)[:-19] for f in files])
                self.filename = ''.join([f+',' for f in files])[:-1]
                self.filedate = datetime.strptime(
                    basename(files[0])[-18:-10]+' 00:00:00', '%Y%m%d %H:%M:%S'
                    ).replace(tzinfo=timezone.utc)

                # establish time attributes from filenames
                for p in patterns:
                    # get list of files to loop through later
                    pattern_files = sorted(glob(file_dir+p+'*.nc'))
                    h0_file = [f for f in pattern_files if 'h0' in f]
                    if len(h0_file) > 0:
                        pattern_files.remove(h0_file[0])  # should be only one
                    self.pattern_files[p] = pattern_files
                    self.times[p] = {'start': [], 'end': [], 'all': []}

                    # loop through to get times, one file per time
                    time_str = [basename(f)[-18:-3] for f in pattern_files]
                    times = RU.str_to_hrs(time_str, self.filedate,
                                          format_string='%Y%m%d_%H%M%S')
                    self.times[p]['start'] = array(times)
                    self.times[p]['end'] = array(times)
                    self.times[p]['all'] = array(times)

                # create time list file if DNE
                RU.create_timelist(list_file, time_file, self.modelname,
                                   self.times, self.pattern_files,
                                   self.filedate)
            else:  # read in data and time grids from file list
                self.times, self.pattern_files, self.filedate, self.filename =\
                    RU.read_timelist(time_file, list_file)
            if filetime:
                return  # return times as is to prevent infinite recursion

            # store variables
            self.missing_value = NaN
            self.varfiles = {}  # store which variable came from which file
            self.gvarfiles = {}  # store file variable name similarly
            self.err_list = []
            self.variables = {}

            if printfiles:
                print(f'{len(self.filename)} Files:')
                for file in self.filename:
                    print(file)

            # perform initial check on variables_requested list
            if len(variables_requested) > 0 and variables_requested != 'all':
                test_list = [value[0] for key, value in model_varnames.items()]
                err_list = [item for item in variables_requested if item not in
                            test_list]
                if len(err_list) > 0:
                    print('Variable name(s) not recognized:', err_list)
                for item in err_list:
                    variables_requested.remove(item)
                if len(variables_requested) == 0:
                    return

            # collect variable dependency lists
            self.total_ilev = [value[0] for key, value in
                               model_varnames.items() if value[-2][-1] ==
                               'ilev' and value[0] != 'H_ilev']
            self.total_replace = [item.split('_ilev')[0] for item in
                                  self.total_ilev if item != 'H_ilev']
            # dictionary mapping to navigate related variable names
            self.ilev_map = {item1: item2 for item1, item2 in
                             zip(self.total_replace, self.total_ilev)}
            # determine variable mapping
            if len(variables_requested) > 0 and variables_requested != 'all':
                # add ilev version of variables to the list, adding H_ilev
                add_ilev = [var+'_ilev' for var in variables_requested if var
                            in self.ilev_map.keys()]
                if len(add_ilev) > 0:  # add to list
                    add_ilev += ['H_ilev']  # add for plev inversion
                new_var = variables_requested + add_ilev
                short_var = [item for item in new_var if item not
                             in self.ilev_map.keys()]  # remove replaced items

            # collect variable names and coordinate grids per pattern type
            self.varfiles, self.gvarfiles = {}, {}
            self.err_list, self.var_dict = [], {}
            for p in self.pattern_files.keys():
                # check var_list for variables not possible in this file set
                cdf_data = Dataset(self.pattern_files[p][0], 'r')
                if len(variables_requested) > 0 and \
                        variables_requested != 'all':
                    gvar_list = [key for key, value in model_varnames.items()
                                 if value[0] in short_var and key in
                                 cdf_data.variables.keys()]  # file var names
                    # check for variables requested but not available in p file
                    if len(gvar_list) != len(short_var):
                        err_list = [value[0] for key, value in
                                    model_varnames.items() if value[0] in
                                    short_var and key not in
                                    cdf_data.variables.keys()]
                        self.err_list.extend(err_list)  # add to master list
                else:
                    gvar_list = [key for key in model_varnames.keys()
                                 if key in cdf_data.variables.keys()]
                # store which file these variables came from
                self.varfiles[p] = [model_varnames[key][0] for key in
                                    gvar_list]
                self.gvarfiles[p] = gvar_list

                # collect all possible variables in set of files and return
                if variables_requested == 'all':
                    var_dict = {value[0]: value[1:] for key, value in
                                model_varnames.items() if key in gvar_list}
                    # add non-ilev versions of the variables in the files
                    key_list = list(var_dict.keys())
                    for var_key in key_list:
                        self.var_dict[var_key] = var_dict[var_key]
                        if var_key in self.total_ilev:
                            # retrieve equivalent non-ilev variable name
                            new_key = [key for key, value in
                                       self.ilev_map.items() if value ==
                                       var_key][0]
                            # retrieve key for model_varnames mapping
                            model_key = [key for key, value in
                                         model_varnames.items() if
                                         value[0] == new_key][0]
                            # add to returned dictionary
                            self.var_dict[model_varnames[model_key][0]] =\
                                model_varnames[model_key][1:]

                # retrieve coordinate grids, assuming constant in time
                lon = array(cdf_data.variables['lon'])
                lon_le180 = list(where(lon <= 180)[0])  # 0 to 180
                lon_ge180 = list(where(lon >= 180)[0])
                tmp = append(lon, 360.) - 180.  # now -180. to +180.
                setattr(self, '_lon_'+p, tmp)
                setattr(self, '_lon_idx_'+p, lon_ge180+lon_le180)
                setattr(self, '_lat_'+p,
                        array(cdf_data.variables['lat']))
                if 'alt' in cdf_data.variables.keys():
                    setattr(self, '_height_'+p,
                            array(cdf_data.variables['alt']))  # km
                    setattr(self, '_heightunits_'+p, 'km')
                # determine if pressure leve is needed (not included)
                ilev_check = [True if 'ilev' in var else False for var in
                              self.varfiles[p]]
                if sum(ilev_check) > 0:
                    # create a pressure level coordinate grid
                    idx = ilev_check.index(True)  # get first index of 3D var
                    varname = self.gvarfiles[p][idx]
                    grid_length = cdf_data.variables[varname].shape[0]
                    setattr(self, '_ilev_'+p,
                            linspace(1, grid_length, grid_length))

                    # get median km grid from h0 file
                    if 'height' in cdf_data.variables.keys():
                        h0_file = self.pattern_files[p][0][:-18] + 'h0.nc'
                        if isfile(h0_file):
                            cdf_h = Dataset(h0_file)
                            self._km_ilev = array(cdf_h.variables['km_ilev'])
                            self._km_ilev_max = cdf_h.km_max
                            self._km_ilev_min = cdf_h.km_min
                            cdf_h.close()
                cdf_data.close()
            # return var_dict (already created)
            if variables_requested == 'all':
                return

            # loop through patterns to initialize variables dictionary
            for p in self.pattern_files.keys():
                gvar_list = self.gvarfiles[p]
                variables = {model_varnames[key][0]: {
                    'units': model_varnames[key][-1], 'data': p}
                    for key in gvar_list}
                for key in variables.keys():
                    self.variables[key] = variables[key]

            # remove successful variables from err_list
            self.err_list = list(unique(self.err_list))
            self.err_list = [item for item in self.err_list if item not in
                             self.variables.keys()]
            if len(self.err_list) > 0:
                print('Some requested variables are not available in the ' +
                      'files found:\n', self.err_list)

            # register interpolators for each requested variable
            # rearrange to deal with H_ilev and H_ilev1 first if there
            varname_list = list(self.variables.keys())
            if 'H_ilev' in varname_list:
                varname_list.remove('H_ilev')
                varname_list = ['H_ilev'] + varname_list
            t_reg = perf_counter()
            for varname in varname_list:
                self.register_variable(varname, gridded_int)
            if verbose:
                print(f'Took {perf_counter() - t_reg:.5f}s to register ' +
                      f'{len(varname_list)} variables.')
            if verbose:
                print(f'Took a total of {perf_counter() - t0:.5f}s to ' +
                      f'kamodofy {len(varname_list)} variables.')
            return

        # define and register a 3D variable
        def register_variable(self, varname, gridded_int):
            """Registers an interpolator with proper signature"""

            # determine coordinate variables and xvec by coord list
            key = self.variables[varname]['data']
            coord_list = [value[5] for key, value in model_varnames.items()
                          if value[0] == varname][0]
            coord_dict = {'time': {'units': 'hr',
                                   'data': self.times[key]['all']}}
            gvar = [key for key, value in model_varnames.items() if
                    value[0] == varname][0]
            # retrieve coordinate grids
            if 'lat' in coord_list:
                coord_dict['lon'] = {'units': 'deg', 'data':
                                     getattr(self, '_lon_'+key)}
                coord_dict['lat'] = {'units': 'deg', 'data':
                                     getattr(self, '_lat_'+key)}
                lon_idx = getattr(self, '_lon_idx_'+key)
            if 'ilev' in coord_list:
                try:
                    coord_dict['ilev'] = {'units': 'm/m', 'data':
                                          getattr(self, '_ilev_'+key)}
                    if hasattr(self, '_ilevunits_'+key):
                        coord_dict['ilev']['units'] = \
                            getattr(self, '_ilevunits_'+key)
                except:
                    coord_list.remove('ilev')
                    coord_list.append('height')
                    new_varname = varname.split('_ilev')[0]
                    self.variables[new_varname] = self.variables[varname]
                    varname = new_varname
                    if gvar[-2:] == '_2':
                        gvar = gvar[:-2]
            if 'height' in coord_list:
                coord_dict['height'] = {'units': 'km', 'data':
                                        getattr(self, '_height_'+key)}
                if hasattr(self, '_heightunits_'+key):
                    coord_dict['height']['units'] = \
                        getattr(self, '_heightunits_'+key)
            coord_str = [value[3]+value[4] for key, value in
                         model_varnames.items() if value[0] == varname][0]

            # define operations for each variable when given the key
            def func(i):
                '''i is the time slice. WAM-IPE has one time slice per file.'''
                # get data from file
                file = self.pattern_files[key][i]
                cdf_data = Dataset(file)
                data = array(cdf_data.variables[gvar])
                # data wrangling
                if hasattr(cdf_data.variables[gvar], '_FillValue'):
                    fill_value = cdf_data.variables[gvar]._FillValue
                    if fill_value in data:
                        data = where(data == fill_value, NaN, data)
                cdf_data.close()
                return data.T[lon_idx]

            # define and register the interpolators
            # need H functions to be gridded regardless of gridded_int value
            if varname == 'H_ilev':
                gridded_int = True
            self = RU.Functionalize_Dataset(
                self, coord_dict, varname, self.variables[varname],
                gridded_int, coord_str, interp_flag=1, func=func)

            # create pressure level -> km function once per ilev type
            if (varname == 'H_ilev' or 'ilev' in coord_list) and (
                    hasattr(self, '_km_ilev')):  # only if inversion is possibl
                if varname == 'H_ilev':  # create custom interp
                    new_varname = 'Plev'
                    # perform unit conversion if needed
                    if self.variables[varname]['units'] != 'km':
                        self[varname+'km_ijk[km]'] = varname + '_ijk'
                        km_interp = self[varname+'km_ijk']
                    else:
                        km_interp = self[varname+'_ijk']
                    # Import and call custom interpolator
                    units = coord_dict['ilev']['units']
                    self[new_varname], interp_ijk = RU.PLevelInterp(
                        km_interp, coord_dict['time']['data'],
                        coord_dict['lon']['data'], coord_dict['lat']['data'],
                        coord_dict['ilev']['data'], units, self._km_ilev,
                        [self._km_ilev_min, self._km_ilev_max])
                    # kms is a 1D array of the median height values in km
                else:  # define by function composition
                    new_varname = varname.split('_ilev')[0]
                    units = self.variables[varname]['units']
                    self[new_varname] = varname+'(Plev)'
                    interp_ijk = self[new_varname]
                    self[new_varname].meta['arg_units'] = \
                        self['Plev'].meta['arg_units']
                self.variables[new_varname] = {'units': units, 'data': key}

                # create gridded interpolator if requested
                if gridded_int:
                    self = RU.register_griddedPlev(
                        self, new_varname, units, interp_ijk, coord_dict,
                        self._km_ilev, [self._km_ilev_min, self._km_ilev_max])
            return
    return MODEL



# -*- coding: utf-8 -*-
"""
Created on Tue Jun  8 10:18:13 2021

@author: rringuet
version 3: dimension agnostic
version 3a: separated offsets, adapted to new SF structure
version 3b: update to new version of SF code, enable time reconstruction
    options.
version 3c: fine-tune code based on testing, add nan feature
version 3d: add averaging of model slices over two dimensions not reconstructed
version 4: implement coordinate-dependent plot labeling
Public version 0: fine-tuned orbit slicing to produce correct answer by
    subtracting longitude per time (15 deg/hr) for imaginary satellites in
    model orbit slicing to avoid shifting the local time of the new
    'satellites'.
"""

import numpy as np
import time as ti
from csv import DictReader
from datetime import datetime, timezone
from kamodo import Kamodo, kamodofy, gridify
import kamodo_ccmc.flythrough.SatelliteFlythrough as SF
from scipy.interpolate import RegularGridInterpolator
import kamodo_ccmc.flythrough.model_wrapper as MW


def read_GDC_sattraj(filename):
    # read data from Bob's file into dictionary
    file_dict = {}
    sat_file = DictReader(open(filename, 'r'), delimiter='\t')
    for row in sat_file:
        for column, value in row.items():
            file_dict.setdefault(column, []).append(float(value))
    for key in file_dict.keys():
        if key in ['Year', 'Month', 'Day', 'Hour', 'Minute', 'Second',
                   'Millisecond']:
            file_dict[key] = np.array(file_dict[key], dtype=int)
        else:
            file_dict[key] = np.array(file_dict[key], dtype=float)

    # convert time data into utc timestamps (millisec all equal 0.0, ignore)
    file_dict['UTCtimestamps'] = time_to_utcts(
        file_dict['Year'], file_dict['Month'], file_dict['Day'],
        file_dict['Hour'], file_dict['Minute'], file_dict['Second'])
    return file_dict


@np.vectorize
def time_to_utcts(year, month, day, hour, minute, second):
    '''converts given values into a utc timestamp'''
    dt = datetime(year, month, day, hour, minute, second, tzinfo=timezone.utc)
    return datetime.timestamp(dt)


def ts_to_dt(time_val):
    '''Convert UTC timestamp to a UTC datetime object.'''
    return datetime.utcfromtimestamp(time_val).replace(tzinfo=timezone.utc)


@np.vectorize
def ts_to_hours(time_val):
    '''return the utc hour from a utc timestamp.'''
    return datetime.utcfromtimestamp(time_val).hour


def ts_to_LT(time_val, longitude):
    '''Converts the utc timestamp into a local time using longitude.'''
    # convert to local time
    hours = ts_to_hours(time_val)
    LT = np.array(np.floor(longitude/15.), dtype=int) + hours

    # correct local times to be from 0 to 23
    neg_idx = np.where(LT < 0)[0]
    LT[neg_idx] += 24
    high_idx = np.where(LT > 23)[0]
    LT[high_idx] -= 24
    return LT


class Constellation(Kamodo):
    '''Uses shifted copies of the satellite trajectory to reconstruct
    the model data for the variable name given. See
    https://doi.org/10.22541/essoar.167214257.73153757/v1 for more info.

    - model: 'CTIPe', 'IRI', 'GITM', 'SWMF_IE', 'TIEGCM', etc
        (See MW.Choose_Model() output.)
    - variable_name = choose from list of standardized variable names from
        model chosen
    - file_dir = file path to model data
    - sat_time: a numpy array of the utc timestamps
    - c1, c2, c3: numpy arrays of the positions correlating to the utc
        timestamps (c1, c2, c3) should be (x,y,z) in R_E for cartesian
        coordinates, and (lon, lat, radius (R_E) or altitude (km)) for
        spherical coordinates.
    - coord_type: one of 'GDZ', 'GEO', 'GSM', 'GSE', 'SM', 'GEI', 'MAG', 'SPH',
        'RLL' or those supported by AstroPy (See flythrough.utils.CoordConv)
    - coord_grid: either 'car' or 'sph'. Note that not all combinations
        make sense (e.g. 'SPH' and 'car') and are not allowed.
        All cartesian coordinate systems expect x, y, z values
        in R_E. All times should be in UTC timestamps.
        Note: If the input coordinates are spherical and you include offsets in
            latitude, make sure the offsets in latitude do not result in
            latitudes outside of the typical (-90,90) range in degrees. The
            program cannot compensate and will error.
    - recon_option = string representing reconstruction and comparison method
        Choose from 'UnMod_AvgSlice','UnMod_AvgDSlice','AvgMod_AvgSlice',
        'AvgMod_AvgDSlice', 'UnMod_OrbitSliceD', 'UnMod_OrbitSliceN',
        'AvgMod_OrbitSliceD', and 'AvgMod_OrbitSliceN'
        - UnMod means the satellite trajectory is not modified
        - AvgMod means the dimensions not reconstructed are replaced with
          either the average value of the original and offset satellite
          trajectories or the value given in the xx_avg keywords.
        - AvgSlice means recon is compared with a slice at the average
          dimensions not reconstructed, again using either the average or the
          given value in the xx_avg keywords.
        - AvgDSlice means recon is compared with an average of all the slices
          in the range of the dimensions not reconstructed. Weighted averages
          are not supported.
        - OrbitSlice means the model data is retrieved along all trajectories
          in the given constellation instead of along a given dimensional
          average.
          The OrbitSlice option is what an inifite number of satellites would
          see if positioned along the trajectory with time offsets t+dt, where
          dt is the difference in time between satellites along the orbit. For
          computation purposes, the user may enter the value of dt in seconds
          in the input line or use the default of 60 seconds. The trajectories
          are split into orbits, defined as where the latitude differential
          changes from positive to negative, and then extracted from the model
          for the calculated time grid. The 'D' and 'N' variations are for
          retrieving only the day or night portions of the orbit rather than
          the entire orbit to avoid averaging data of different natures.
          Whether a given time+spatial coordinate is on the day or night side
          is determined by converting the time+spatial coordinates of the orbit
          slices into GSE spherical coordinates and filtering based on
          longitude. The longitude logic compensates to keep equatorial
          crossing in GSE at the same MLT as the original set.
          NOTE: Orbital slicing may not produce the desired result in cartesian
          coordinates.
    - recon_dims = string representing choice of dimensions to use in
        reconstruction. 'tc1','tc2','tc3','c1c2','c1c3','c2c3' are the options.
        The first name is the x-axis, and the second name is the y-axis.
    - time_offsets: a list of time offsets in seconds to be applied to the
        trajectories (in s)
    - c1_offsets: a list of offsets to be applied to the c1 coordinate in the
        trajectory in the same units indicated by the coordinate system options
    - c2_offsets: a list of offsets to be applied to the c2 coordinate in the
        trajectory in the same units indicated by the coordinate system options
    - c3_offsets: a list of offsets to be applied to the c3 coordinate in the
        trajectory in the same units indicated by the coordinate system options
    - the offset lists entered MUST BE THE SAME LENGTH or be equal to [0.], and
        must be in the same coordinate system and units as the input
        coordinates.
    - t_avg: the average time in UTC seconds used for when time is not a
        reconstructed dimension and the recon_option chosen requires averaging.
        Default is the average time of the entire set of input and offset
        trajectories.
    - c1_avg: the average c1 value used for when c1 is not a reconstructed
        dimension and the recon_option chosen requires averaging. Default is
        the average c1 value of the entire set of input and offset
        trajectories.
    - c2_avg: the average c2 value used for when c2 is not a reconstructed
        dimension and the recon_option chosen requires averaging. Default is
        the average c2 value of the entire set of input and offset
        trajectories.
    - c3_avg: the average c3 value used for when c3 is not a reconstructed
        dimension and the recon_option chosen requires averaging. Default is
        the average c3 value of the entire set of input and offset
        trajectories.
    - dx = grid resolution for the x axis in the same units as x, default of 2.
    - dy = grid resolution for the y axis in the same units as y, default of 2.
    - d1 = grid resolution for the first non-reconstructed dimension. Only used
        if *_AvgDSlice options are chosen and if slicing in local time.
        Default=0.
    - d2 = grid resolution for the second non-reconstructed dimension. Only
        used if *_AvgDSlice options are chosen. Default=0.
    - dt = time resolution of sampling for orbit slicing option. Default=60 s.
    - run_option = one of 'all' or 'flythrough'. Default is all.
    Note: The dx and dy values are assumed to be in the units of the associated
        coordinate indicated by the coordinate options (s, R_E, deg, or km).
    Note: When reconstructing in two spatial cartesian coordinates
        (e.g. xy, yz, or xz), filter the input satellite trajectory to only
        include times where the third spatial dimension inputs are either
        positive OR negative values to avoid a combined reconstruction of both
        hemispheres in the same plots. For example, if you wish to perform a
        reconstruction of the x and y dimensions in spatial coordinates for
        the North Pole, then filter the satellite trajectory to only include
        times where the z dimension is positive. Similarly, reconstructions of
        the South Pole can be performed by simply filtering out the times where
        the z dimension is positive. This filtering is not needed when time is
        one of the reconstructed dimensions, but it can be done if desired.
        Example:
            import numpy as np
            pos_idx = np.where(results['c3']>0.)[0]  # North Pole is z>0.
            time = results['utc_time'][pos_idx]
            c1 = results['c1'][pos_idx]
            c2 = results['c2'][pos_idx]
            c3 = results['c3'][pos_idx]
    Cartesian note: For cartesian reconstructions of nearly spherical
        trajectories, 'Unmod_AvgDSlice' option is recommended to avoid
        reconstructing a ring of data around the Earth. Reconstructions will be
        better achieved in spherical coordinates for these scenarios. To easily
        switch between coordinate systems, from GSE cartesian to GEO spherical,
        for example:
            from kamodo_ccmc.flythrough.utils import ConvertCoord
            lon, lat, radius, units = ConvertCoord(utc_time, x, y, z,
                          'GSE', 'car', 'GEO', 'sph')
        See the relevant flythrough example notebook for more details on this
        function.
    '''

    def __init__(self, model, variable_name, file_dir, sat_time, c1,
                 c2, c3, coord_type, coord_grid, recon_option, recon_dims,
                 time_offsets=[0.], c1_offsets=[0.], c2_offsets=[0.],
                 c3_offsets=[0.], dx=2., dy=2., d1=0., d2=0., LT='',
                 t_avg='', c1_avg='', c2_avg='', c3_avg='', dt=60.,
                 run_option='all', **kwargs):
        super(Constellation, self).__init__(**kwargs)
        t_start = ti.perf_counter()  # Begin execution timer

        # store input values
        self.model = model
        self.variable_name = variable_name
        self.file_dir = file_dir  # file path to model data
        self.dx, self.dy = dx, dy  # grid resolution for x and y axes
        self.d1, self.d2 = d1, d2  # resolution for non-reconstructed dims
        # save input satellite trajectory time (utc timestamps)
        self.sat_time = sat_time
        self.c1, self.c2, self.c3 = c1, c2, c3  # save input spatial trajectory
        self.coord_type = coord_type
        self.coord_grid = coord_grid  # 'car' or 'sph'
        self.recon_option = recon_option
        self.recon_dims = recon_dims  # which two dimensions to reconstruct
        # datetime object for first utc timestamp
        self.sat_datetime = ts_to_dt(sat_time[0])
        self.t_avg, self.c1_avg, self.c2_avg, self.c3_avg = t_avg, c1_avg,\
            c2_avg, c3_avg
        self.dt = dt  # used for orbit slicing
        if 't' in recon_dims:
            self.dx = dx/3600.  # change dx from sec to hours
        self.run_option = run_option  # either only flythrough or all

        # check inputs for validity
        if recon_option not in ['UnMod_AvgSlice', 'UnMod_AvgDSlice',
                                'AvgMod_AvgSlice', 'AvgMod_AvgDSlice',
                                'UnMod_OrbitSliceD', 'UnMod_OrbitSliceN',
                                'AvgMod_OrbitSliceD', 'AvgMod_OrbitSlice_N']:
            print("Possible values for recon_option are 'UnMod_AvgSlice', " +
                  "'UnMod_AvgDSlice', 'AvgMod_AvgSlice', 'AvgMod_AvgDSlice'," +
                  " 'UnMod_OrbitSliceD', 'UnMod_OrbitSliceN', " +
                  "'AvgMod_OrbitSliceD', and 'AvgMod_OrbitSlice_N'.")
            raise AttributeError('Value for recon_option not recognized.')
        if recon_dims not in ['tc1', 'tc2', 'tc3', 'c1c2', 'c1c3', 'c2c3']:
            print("Possible values for recon_dims are 'tc1', 'tc2', 'tc3', " +
                  "'c1c2', 'c1c3', or 'c2c3'.")
            raise AttributeError('Value for recon_dims not recognized.')
        if 'AvgDSlice' in recon_option and (d1 == 0. or d2 == 0.):
            raise AttributeError('You must specify d1 and d2 for ' +
                                 f'{self.recon_option}.')

        # sometimes see [0.,0.,0.] as an offset for no input values (?!).
        # seems to be a memory random error. Doesn't happen all the time.
        if sum(time_offsets) == 0:
            time_offsets = [0.]
        if sum(c1_offsets) == 0:
            c1_offsets = [0.]
        if sum(c2_offsets) == 0:
            c2_offsets = [0.]
        if sum(c3_offsets) == 0:
            c3_offsets = [0.]

        # check that offset lists are the same length. append zeros if not
        # print(time_offsets, c1_offsets, c2_offsets, c3_offsets)
        max_length = max([len(item) for item in [time_offsets, c1_offsets,
                                                 c2_offsets, c3_offsets]])
        for item in [time_offsets, c1_offsets, c2_offsets, c3_offsets]:
            while len(item) < max_length:
                item.append(0.)
        self.time_offsets = time_offsets
        self.c1_offsets = c1_offsets
        self.c2_offsets = c2_offsets
        self.c3_offsets = c3_offsets

        # get variable units from satellite flythrough software
        self.variable_units = MW.Var_units(model,
                                           [variable_name])[variable_name]
        self.variables = {}   # initialize kamodo's dictionary

        # BEGIN ANALYSIS
        if self.run_option == 'all':
            # fly through model data, accounting for shifts and wrapping of lon
            results = self.flythrough()  # full dictionary from flythrough
            # sort data onto a grid, assign interpolator
            self.recon_grid = self.grid_data(results)
            self.create_2Dinterpolator(self.variable_name, self.x, self.y,
                                       self.recon_grid, self.variable_units)

            # retrieve model data on identical grid, assign interpolators
            self.model_data = self.model_grid()
            self.create_2Dinterpolator(self.variable_name+'_model', self.x,
                                       self.y, self.model_data,
                                       self.variable_units)
            pdiff = (self.model_data - self.recon_grid)/self.model_data * 100.
            self.create_2Dinterpolator('PercentDiff', self.x, self.y,
                                       pdiff, '')
        elif self.run_option == 'flythrough':
            # fly through model data, accounting for shifts and wrapping of lon
            results = self.flythrough()  # full dictionary from flythrough
            # sort data onto a grid, assign interpolator
            self.recon_grid = self.grid_data(results)
            self.create_2Dinterpolator(self.variable_name, self.x, self.y,
                                       self.recon_grid, self.variable_units)
        else:
            raise AttributeError(f'Run option unknown: {self.run_option}.')
        print('Reconstruction program complete in ' +
              f'{ti.perf_counter()-t_start:.5f} s.')
        return

    # ---------------------- begin functions ----------------------------------

    def flythrough(self):
        '''fly trajectories through model data.'''

        # determine max value of longitude from coordinate type if spherical
        if self.coord_grid == 'sph' and self.coord_type == 'SPH':
            self.lon_max = 360.
        elif self.coord_grid == 'sph' and self.coord_type != 'SPH':
            self.lon_max = 180.

        # calculate offset trajectories and assemble into one array per coord
        fly_t, fly_c1, fly_c2, fly_c3 = [], [], [], []
        for i in range(len(self.time_offsets)):
            # calculate new trajectory
            new_time = self.sat_time + self.time_offsets[i]
            new_c1 = self.c1 + self.c1_offsets[i]
            new_c2 = self.c2 + self.c2_offsets[i]
            new_c3 = self.c3 + self.c3_offsets[i]

            # perform corrections and checks for lon and lat if in sph coords
            if self.coord_grid == 'sph':
                if self.c2_offsets[i] > 0 and (new_c2.min() < -90. or
                                               new_c2.max() > 90.):
                    print('Cannot correct for latitude shifts resulting in ' +
                          'latitudes greater than 90 degrees or less than ' +
                          '-90 degrees. Consider decreasing or removing the ' +
                          f'latitude offset {self.c2_offsets[i]}.')

                # check for and perform any longitude wrapping corrections
                while new_c1.max() > self.lon_max:
                    idx = np.where(new_c1 > self.lon_max)[0]
                    new_c1[idx] -= 360.

            # append arrays to list
            fly_t.append(new_time)
            fly_c1.append(new_c1)
            fly_c2.append(new_c2)
            fly_c3.append(new_c3)

        # change into 1D arrays
        fly_t = np.ravel(np.array(fly_t))
        fly_c1 = np.ravel(np.array(fly_c1))
        fly_c2 = np.ravel(np.array(fly_c2))
        fly_c3 = np.ravel(np.array(fly_c3))
        self.fly_t, self.fly_c1, self.fly_c2, self.fly_c3 = fly_t, fly_c1,\
            fly_c2, fly_c3
        ll = len(fly_t)

        # store average of entire trajectory if needed next or later or both
        if 'Avg' in self.recon_option:
            if self.t_avg == '':
                self.t_avg = np.mean(fly_t)
            if self.c1_avg == '':
                self.c1_avg = np.mean(fly_c1)
            if self.c2_avg == '':
                self.c2_avg = np.mean(fly_c2)
            if self.c3_avg == '':
                self.c3_avg = np.mean(fly_c3)

        # if AvgMod_AvgSlice is chosen, calculate avg of dims not reconstructed
        if self.recon_option[:6] == 'AvgMod':
            if 't' not in self.recon_dims:
                fly_t = np.repeat(self.t_avg, ll)
                print('Average time used for trajectories:', self.t_avg)
            if 'c1' not in self.recon_dims:
                fly_c1 = np.repeat(self.c1_avg, ll)
                print('Average c1 used for trajectories:', self.c1_avg)
            if 'c2' not in self.recon_dims:
                fly_c2 = np.repeat(self.c2_avg, ll)
                print('Average c2 used for trajectories:', self.c2_avg)
            if 'c3' not in self.recon_dims:
                fly_c3 = np.repeat(self.c3_avg, ll)
                print('Average c3 used for trajectories:', self.c3_avg)

        if 'OrbitSlice' in self.recon_option:
            # retrieve local times through SM coordinate conversion
            print('Filtering constellation trajectories on day/night option ' +
                  f'using GSE cartesian coordinates for {fly_t.size} ' +
                  'locations...')
            from kamodo_ccmc.flythrough.utils import ConvertCoord
            c1_GSE, c2_GSE, c3_GSE, units = ConvertCoord(
                fly_t, fly_c1, fly_c2, fly_c3, self.coord_type,
                self.coord_grid, 'GSE', 'car')

            # split into day or night based on option chosen
            # In GSE sph coordinates, day is lon>90 or lon<-90
            if 'OrbitSliceD' in self.recon_option:
                LT_idx = np.where(c1_GSE >= 0.)[0]  # x points to sun
            elif 'OrbitSliceN' in self.recon_option:
                LT_idx = np.where(c1_GSE <= 0.)[0]
            fly_t, fly_c1 = fly_t[LT_idx], fly_c1[LT_idx]
            fly_c2, fly_c3 = fly_c2[LT_idx], fly_c3[LT_idx]

        # decrease computation time by removing times not in files
        beg, end = MW.File_Times(self.model, self.file_dir, print_output=False)
        beg_ts, end_ts = beg.timestamp(), end.timestamp()
        tidx = np.where((fly_t >= beg_ts) & (fly_t <= end_ts))[0]
        if len(tidx) > 0:
            if len(tidx) < len(fly_t):
                diff = len(fly_t) - len(tidx)
                print(f'{diff} times not in model data. Removing.')
            fly_t, fly_c1 = fly_t[tidx], fly_c1[tidx]
            fly_c2, fly_c3 = fly_c2[tidx], fly_c3[tidx]

        # perform flythrough
        print('Performing satellite constellation flythrough for ' +
              f'{fly_t.size} locations...')
        results = SF.ModelFlythrough(
            self.model, self.file_dir, [self.variable_name], fly_t, fly_c1,
            fly_c2, fly_c3, self.coord_type+'-'+self.coord_grid,
            _print_units=False)

        # save correct time array for flythrough (all bad times removed)
        self.time_min = results['utc_time'].min()
        self.time_max = results['utc_time'].max()
        if self.t_avg > self.time_max or self.t_avg < self.time_min:
            self.t_avg = np.mean([self.time_min, self.time_max])
            print('Average time not in model data. Choosing new time in ' +
                  f'middle of model data time range: {self.t_avg}s or ' +
                  f'{ts_to_dt(self.t_avg)}')
        self.data_datetime = ts_to_dt(results['utc_time'][0])
        # self.results = results  # too much demand on memory
        return results

    def grid_data(self, results):
        '''Grid data along axes indicated. Handles all six options.'''
        t0 = ti.perf_counter()

        # collect variable data from satellite flythrough results
        variable_data = results[self.variable_name]

        # perform specialized gridding for sph coords (if lat or lon involved)
        if self.recon_dims == 'c1c2' and self.coord_grid == 'sph':  # 'LonLat'
            # initialize data
            x_data = results['c1']
            y_data = results['c2']

            # define grid if not already defined
            if not (hasattr(self, 'x') and hasattr(self, 'y')):
                # create lon, lat coordinate grid
                self.xmin, self.xmax = x_data.min(), x_data.max()
                self.ymin, self.ymax = y_data.min(), y_data.max()
                self.nx = int(360./self.dx) + 1
                self.ny = int(180./self.dy) + 1
                # center positions
                self.x = np.linspace(self.lon_max-360., self.lon_max, self.nx)
                self.y = np.linspace(-90., 90., self.ny)
            else:
                print('Using predefined grid.')
            variable = np.zeros((self.nx, self.ny))

            # sort data into grid
            print('Gridding data for LonLat...', end="")
            for i in range(self.nx):
                for j in range(self.ny):
                    idx = np.where((abs(x_data-self.x[i]) < self.dx/2.) &
                                   (abs(y_data-self.y[j]) < self.dy/2.))[0]
                    if len(idx) > 0:
                        variable[i, j] = np.nanmean(variable_data[idx])
                    else:
                        variable[i, j] = np.nan

        elif self.recon_dims == 'c2c3' and self.coord_grid == 'sph':
            # initialize data and height boundary
            x_data = results['c2']
            y_data = results['c3']

            # define grid if not already defined
            if not (hasattr(self, 'x') and hasattr(self, 'y')):
                # create lat, height coordinate grid
                self.xmin, self.xmax = x_data.min(), x_data.max()
                self.ymin, self.ymax = y_data.min(), y_data.max()
                self.nx = int(180./self.dx) + 1
                self.ny = int((y_data.max()-y_data.min())/self.dy) + 1
                self.x = np.linspace(-90., 90., self.nx)  # center positions
                self.y = np.linspace(y_data.min(), y_data.max(), self.ny)
            else:
                print('Using predefined grid.')
            variable = np.zeros((self.nx, self.ny))

            # sort data into grid
            print('Gridding data for LatHR...', end="")
            for i in range(self.nx):
                for j in range(self.ny):
                    idx = np.where((abs(x_data-self.x[i]) < self.dx/2.) &
                                   (abs(y_data-self.y[j]) < self.dy/2.))[0]
                    if len(idx) > 0:
                        variable[i, j] = np.nanmean(variable_data[idx])
                    else:
                        variable[i, j] = np.nan

        elif self.recon_dims == 'c1c3' and self.coord_grid == 'sph':
            # initialize data and height boundary
            x_data = results['c1']
            y_data = results['c3']

            # define grid if not already defined
            if not (hasattr(self, 'x') and hasattr(self, 'y')):
                # create lon, height coordinate grid
                self.xmin, self.xmax = x_data.min(), x_data.max()
                self.ymin, self.ymax = y_data.min(), y_data.max()
                self.nx = int(360./self.dx) + 1
                self.ny = int((y_data.max()-y_data.min())/self.dy) + 1
                # center positions
                self.x = np.linspace(self.lon_max-360., self.lon_max, self.nx)
                self.y = np.linspace(y_data.min(), y_data.max(), self.ny)
            else:
                print('Using predefined grid.')
            variable = np.zeros((self.nx, self.ny))

            # sort data into grid
            print('Gridding data for LonHR...', end="")
            for i in range(self.nx):
                for j in range(self.ny):
                    idx = np.where((abs(x_data-self.x[i]) < self.dx/2.) &
                                   (abs(y_data-self.y[j]) < self.dy/2.))[0]
                    if len(idx) > 0:
                        variable[i, j] = np.nanmean(variable_data[idx])
                    else:
                        variable[i, j] = np.nan

        elif self.recon_dims == 'tc2' and self.coord_grid == 'sph':
            # initialize data and time boundary
            xutc_data = results['utc_time']
            # convert to hours for plotting
            x_data = (xutc_data-xutc_data.min()) / 3600.
            y_data = results['c2']

            # define grid if not already defined
            if not (hasattr(self, 'x') and hasattr(self, 'y')):
                # create time, lat coordinate grid
                self.xmin, self.xmax = x_data.min(), x_data.max()
                self.ymin, self.ymax = y_data.min(), y_data.max()
                self.nx = int((x_data.max()-x_data.min())/self.dx) + 1
                self.ny = int(180./self.dy) + 1
                # center positions
                self.x = np.linspace(x_data.min(), x_data.max(), self.nx)
                self.y = np.linspace(-90., 90., self.ny)
            else:
                print('Using predefined grid.')
            variable = np.zeros((self.nx, self.ny))

            # sort data into grid
            print('Gridding data for TimeLat...', end="")
            for i in range(self.nx):
                for j in range(self.ny):
                    idx = np.where((abs(x_data-self.x[i]) < self.dx/2.) &
                                   (abs(y_data-self.y[j]) < self.dy/2.))[0]
                    if len(idx) > 0:
                        variable[i, j] = np.nanmean(variable_data[idx])
                    else:
                        variable[i, j] = np.nan

        elif self.recon_dims == 'tc1' and self.coord_grid == 'sph':
            # initialize data and time boundary
            xutc_data = results['utc_time']
            # convert to hours for plotting
            x_data = (xutc_data-xutc_data.min()) / 3600.
            y_data = results['c1']

            # define grid if not already defined
            if not (hasattr(self, 'x') and hasattr(self, 'y')):
                # create time, lon coordinate grid
                self.xmin, self.xmax = x_data.min(), x_data.max()
                self.ymin, self.ymax = y_data.min(), y_data.max()
                self.nx = int((x_data.max()-x_data.min())/self.dx) + 1
                self.ny = int(360./self.dy) + 1
                # center positions
                self.x = np.linspace(x_data.min(), x_data.max(), self.nx)
                self.y = np.linspace(self.lon_max-360., self.lon_max, self.ny)
            else:
                print('Using predefined grid.')
            variable = np.zeros((self.nx, self.ny))

            # sort data into grid
            print('Gridding data for TimeLon...', end="")
            for i in range(self.nx):
                for j in range(self.ny):
                    idx = np.where((abs(x_data-self.x[i]) < self.dx/2.) &
                                   (abs(y_data-self.y[j]) < self.dy/2.))[0]
                    if len(idx) > 0:
                        variable[i, j] = np.nanmean(variable_data[idx])
                    else:
                        variable[i, j] = np.nan

        elif 't' in self.recon_dims:  # 'tc3' sph and all car time options
            # initialize data and boundaries
            xutc_data = results['utc_time']
            # convert to hours for plotting
            x_data = (xutc_data-xutc_data.min()) / 3600.
            y_name = self.recon_dims[-2:]
            y_data = results[y_name]

            # define grid if not already defined
            if not (hasattr(self, 'x') and hasattr(self, 'y')):
                # create time, height coordinate grid
                self.xmin, self.xmax = x_data.min(), x_data.max()
                self.ymin, self.ymax = y_data.min(), y_data.max()
                self.nx = int((x_data.max()-x_data.min())/self.dx) + 1
                self.ny = int((y_data.max()-y_data.min())/self.dy) + 1
                # center positions
                self.x = np.linspace(x_data.min(), x_data.max(), self.nx)
                self.y = np.linspace(y_data.min(), y_data.max(), self.ny)
            else:
                print('Using predefined grid.')
            variable = np.zeros((self.nx, self.ny))

            # sort data into grid
            print('Gridding data for TimeHR...', end="")
            for i in range(self.nx):
                for j in range(self.ny):
                    idx = np.where((abs(x_data-self.x[i]) < self.dx/2.) &
                                   (abs(y_data-self.y[j]) < self.dy/2.))[0]
                    if len(idx) > 0:
                        variable[i, j] = np.nanmean(variable_data[idx])
                    else:
                        variable[i, j] = np.nan

        else:  # cartesian
            # initialize data and boundaries
            x_name = self.recon_dims[:-2]
            x_data = results[x_name]
            y_name = self.recon_dims[-2:]
            y_data = results[y_name]

            # define grid if not already defined
            if not (hasattr(self, 'x') and hasattr(self, 'y')):
                # create coordinate grid
                self.xmin, self.xmax = x_data.min(), x_data.max()
                self.ymin, self.ymax = y_data.min(), y_data.max()
                self.nx = int((x_data.max()-x_data.min())/self.dx) + 1
                self.ny = int((y_data.max()-y_data.min())/self.dy) + 1
                # center positions
                self.x = np.linspace(x_data.min(), x_data.max(), self.nx)
                self.y = np.linspace(y_data.min(), y_data.max(), self.ny)
            else:
                print('Using predefined grid.')
            variable = np.zeros((self.nx, self.ny))

            # sort data into grid
            print('Building cartesian reconstruction grid...', end="")
            for i in range(self.nx):
                for j in range(self.ny):
                    idx = np.where((abs(x_data-self.x[i]) < self.dx/2.) &
                                   (abs(y_data-self.y[j]) < self.dy/2.))[0]
                    if len(idx) > 0:
                        variable[i, j] = np.nanmean(variable_data[idx])
                    else:
                        variable[i, j] = np.nan

        print(f'done in {ti.perf_counter()-t0:.5f}s for {self.nx*self.ny} ' +
              'gridpoints.\n')
        return variable

    def model_grid(self):
        '''Retrieve model grid in method chosen.'''

        # section off code by analysis choice
        # UnMod_AvgSlice, AvgMod_AvgSlice
        if self.recon_option[-8:] == 'AvgSlice':
            # get slice from model data at avg value of each dim not used
            # determine avg value of dimensions not included in reconstruction
            if 't' not in self.recon_dims:
                new_time = self.t_avg
                print('Average t used for model:', new_time)
            if 'c1' not in self.recon_dims:
                new_c1 = self.c1_avg
                print('Average c1 used for model:', new_c1)
            if 'c2' not in self.recon_dims:
                new_c2 = self.c2_avg
                print('Average c2 used for model:', new_c2)
            if 'c3' not in self.recon_dims:
                new_c3 = self.c3_avg
                print('Average c3 used for model:', new_c3)

            # assign values for model grid according to recon_dims chosen
            # x coordinate first
            if self.recon_dims[:-2] == 't':  # convert hrs to UTC timestamps
                new_time = self.x * 3600. + self.time_min  # model_x
            elif self.recon_dims[:-2] == 'c1':
                new_c1 = self.x  # model_x
            elif self.recon_dims[:-2] == 'c2':
                new_c2 = self.x  # model_x
            elif self.recon_dims[:-2] == 'c3':
                new_c3 = self.x  # model_x
            # then y coordinate. the y coordinate is never time
            if self.recon_dims[-2:] == 'c1':
                new_c1 = self.y  # model_y
            elif self.recon_dims[-2:] == 'c2':
                new_c2 = self.y  # model_y
            elif self.recon_dims[-2:] == 'c3':
                new_c3 = self.y  # model_y

            # create coordinate arrays of same length for flythrough
            tt, xx, yy, zz = np.meshgrid(new_time, new_c1, new_c2, new_c3)
            tt, xx, yy, zz = np.ravel(tt), np.ravel(xx), np.ravel(yy),\
                np.ravel(zz)

            # perform flythrough
            print(f'Performing averaged grid flythrough for {tt.size} ' +
                  'locations...')
            t0 = ti.perf_counter()
            model_results = SF.ModelFlythrough(
                self.model, self.file_dir, [self.variable_name], tt, xx, yy,
                zz, self.coord_type+'-'+self.coord_grid, _print_units=False)
            # self.model_results = model_results  # store for sanity checks
            print(f'Grid flythrough completed in {ti.perf_counter()-t0:.5f} ' +
                  's.')

            # reshape data and return. The times should always be defined.
            if self.recon_dims == 'tc1':  # needed only for this combination
                model_data = np.reshape(model_results[self.variable_name],
                                        (self.y.size, self.x.size)).T
            else:
                model_data = np.reshape(model_results[self.variable_name],
                                        (self.x.size, self.y.size))
            return model_data

        # A model orbit slice is what an infinite number of satellites would
        # see if distributed along the orbit slice.
        # The longitude logic here compensates to keep equatorial crossing in
        # GSE at the same MLT as the original set.
        elif 'OrbitSlice' in self.recon_option:
            print('Performing orbit slicing of model data...')
            # determine approx length of an orbit
            lat_diff = np.diff(self.c2)  # find dlat
            idx = np.array([i for i in range(len(lat_diff)-1) if
                            lat_diff[i] > 0 and lat_diff[i+1] < 0],
                           dtype=int) + 2  # where orbits start
            T = []  # list to store orbital periods in for averaging later
            for i in range(len(idx)-1):  # ignore last orbit since partial
                singleorbit_time = self.sat_time[idx[i]: idx[i+1]]  # time vals
                T.append(singleorbit_time.max() - singleorbit_time.min())
            # find average of periods to nearest second
            orbital_period = int(np.mean(np.array(T)))
            print(f'Approximate orbital period is {orbital_period} s.')

            # initialize time offset grid and corresponding lon offsets
            tgrid = np.linspace(0, orbital_period,
                                int(orbital_period/self.dt) + 1)
            time_grid = np.repeat(tgrid, len(self.fly_t))
            lon_grid = -time_grid / 3600. * 15.  # -15 deg lon for each hour

            # apply grid of offsets to full constellation trajectory
            model_t = np.tile(self.fly_t, len(tgrid)) + time_grid
            model_c1 = np.tile(self.fly_c1, len(tgrid)) + lon_grid
            model_c2 = np.tile(self.fly_c2, len(tgrid))
            model_c3 = np.tile(self.fly_c3, len(tgrid))

            # retrieve local times through SM coordinate conversion
            print('Filtering model orbit slices on day/night option using ' +
                  f'GSE cartesian coordinates for {model_t.size} locations...')
            from kamodo_ccmc.flythrough.utils import ConvertCoord
            c1_MAG, c2_MAG, c3_MAG, units = ConvertCoord(
                model_t, model_c1, model_c2, model_c3, self.coord_type,
                self.coord_grid, 'GSE', 'car')

            # split into day or night based on option chosen
            # In GSE sph coordinates, day is lon>90 or lon<-90
            # x points to sun, x>=0 is dayside
            if 'OrbitSliceD' in self.recon_option:
                LT_idx = np.where(c1_MAG >= 0.)[0]
            elif 'OrbitSliceN' in self.recon_option:
                LT_idx = np.where(c1_MAG <= 0.)[0]
            model_t = model_t[LT_idx]
            model_c1 = model_c1[LT_idx]
            model_c2 = model_c2[LT_idx]
            model_c3 = model_c3[LT_idx]

            # perform flythrough
            print(f'{model_t.size} locations remaining for flythrough.')
            t0 = ti.perf_counter()
            model_results = SF.ModelFlythrough(
                self.model, self.file_dir, [self.variable_name], model_t,
                model_c1, model_c2, model_c3,
                self.coord_type+'-'+self.coord_grid, _print_units=False)

            # the orbit slices and day/night filtering result in orbit arrays
            # of different lengths, so can't just reshape. Have to regrid.
            model_data = self.grid_data(model_results)
            return model_data

        else:  # find average of slices in range of dimensions not used
            # assign values for model grid according to recon_dims option
            # chosen. *_AvgDSlice. x coordinate first.
            if self.recon_dims[:-2] == 't':  # convert hrs to UTC timestamps
                new_time = self.x*3600. + self.time_min  # model_x
            elif self.recon_dims[:-2] == 'c1':
                new_c1 = self.x  # model_x
            elif self.recon_dims[:-2] == 'c2':
                new_c2 = self.x  # model_x
            elif self.recon_dims[:-2] == 'c3':
                new_c3 = self.x  # model_x
            # then y coordinate. the y coordinate is never time
            if self.recon_dims[-2:] == 'c1':
                new_c1 = self.y  # model_y
            elif self.recon_dims[-2:] == 'c2':
                new_c2 = self.y  # model_y
            elif self.recon_dims[-2:] == 'c3':
                new_c3 = self.y  # model_y

            # apply d1 and d2 to other two dimensions to decrease computation
            # time. The t and c1 dimensions need to be switched in order for
            # the data to get reshaped correctly. It simply doesn't work the
            # normal way in 4D.
            if self.recon_dims == 'tc1':
                self.n1 = int(np.ceil((self.c2.max()-self.c2.min())/self.d1))
                self.n2 = int(np.ceil((self.c3.max()-self.c3.min())/self.d2))
                new_c2 = np.linspace(self.c2.min(), self.c2.max(), self.n1)
                new_c3 = np.linspace(self.c3.min(), self.c3.max(), self.n2)
                # flip t and c1
                final_shape = (self.ny, self.nx, self.n1, self.n2)
                mean_axes = (2, 3)
            elif self.recon_dims == 'tc2':
                self.n1 = int(np.ceil((self.c1.max()-self.c1.min())/self.d1))
                self.n2 = int(np.ceil((self.c3.max()-self.c3.min())/self.d2))
                new_c1 = np.linspace(self.c1.min(), self.c1.max(), self.n1)
                new_c3 = np.linspace(self.c3.min(), self.c3.max(), self.n2)
                # flip t and c1
                final_shape = (self.n1, self.nx, self.ny, self.n2)
                mean_axes = (1, 3)
            elif self.recon_dims == 'tc3':
                self.n1 = int(np.ceil((self.c1.max()-self.c1.min())/self.d1))
                self.n2 = int(np.ceil((self.c2.max()-self.c2.min())/self.d2))
                new_c1 = np.linspace(self.c1.min(), self.c1.max(), self.n1)
                new_c2 = np.linspace(self.c2.min(), self.c2.max(), self.n2)
                # flip t and c1
                final_shape = (self.n1, self.nx, self.n2, self.ny)
                mean_axes = (1, 2)
            elif self.recon_dims == 'c1c2':
                self.n1 = int(np.ceil((self.time_max-self.time_min)/self.d1))
                self.n2 = int(np.ceil((self.c3.max()-self.c3.min())/self.d2))
                new_time = np.linspace(self.time_min, self.time_max, self.n1)
                new_c3 = np.linspace(self.c3.min(), self.c3.max(), self.n2)
                # have to flip t and c1
                final_shape = (self.nx, self.n1, self.ny, self.n2)
                mean_axes = (0, 3)
            elif self.recon_dims == 'c1c3':
                self.n1 = int(np.ceil((self.time_max-self.time_min)/self.d1))
                self.n2 = int(np.ceil((self.c2.max()-self.c2.min())/self.d2))
                new_time = np.linspace(self.time_min, self.time_max, self.n1)
                new_c2 = np.linspace(self.c2.min(), self.c2.max(), self.n2)
                # flip t and c1
                final_shape = (self.nx, self.n1, self.n2, self.ny)
                mean_axes = (0, 2)
            elif self.recon_dims == 'c2c3':
                self.n1 = int(np.ceil((self.time_max-self.time_min)/self.d1))
                self.n2 = int(np.ceil((self.c1.max()-self.c1.min())/self.d2))
                new_time = np.linspace(self.time_min, self.time_max, self.n1)
                new_c1 = np.linspace(self.c1.min(), self.c1.max(), self.n2)
                # flip t and c1
                final_shape = (self.n2, self.n1, self.nx, self.ny)
                mean_axes = (0, 1)
            final_shape = list(final_shape)

            # create coord arrays of same length, perform flythrough
            tt, xx, yy, zz = np.meshgrid(new_time, new_c1, new_c2, new_c3,
                                         indexing='xy')
            print('Performing non-averaged grid flythrough for ' +
                  f'{np.ravel(tt).size} locations.')
            t0 = ti.perf_counter()
            model_results = SF.ModelFlythrough(
                self.model, self.file_dir, [self.variable_name], np.ravel(tt),
                np.ravel(xx), np.ravel(yy), np.ravel(zz),
                self.coord_type+'-'+self.coord_grid, _print_units=False)
            print(f'Grid flythrough completed in {ti.perf_counter()-t0:.5f} ' +
                  's.')

            # reshape, avg grids over non-reconstructed dimensions and return
            tmp = np.reshape(model_results[self.variable_name],
                             tuple(final_shape))
            model_array = np.transpose(tmp, axes=(1, 0, 2, 3))
            # ignore NaN values in avg
            model_data = np.nanmean(model_array, mean_axes)
            return model_data

    def xvec_dict(self):
        '''determine argument units based on dimension and coordinate choices
        '''

        xvec_dependencies = {}
        if 't' in self.recon_dims:
            xvec_dependencies['Time'] = 'hr'
        if 'c1' in self.recon_dims:
            if self.coord_grid == 'car':
                xvec_dependencies['X'] = 'R_E'
            elif self.coord_grid == 'sph':
                xvec_dependencies['Lon'] = 'deg'
        if 'c2' in self.recon_dims:
            if self.coord_grid == 'car':
                xvec_dependencies['Y'] = 'R_E'
            elif self.coord_grid == 'sph':
                xvec_dependencies['Lat'] = 'deg'
        if 'c3' in self.recon_dims:
            if self.coord_grid == 'car':
                xvec_dependencies['Z'] = 'R_E'
            elif self.coord_grid == 'sph' and self.coord_type == 'GDZ':
                xvec_dependencies['Alt'] = 'km'
            elif self.coord_grid == 'sph' and self.coord_type != 'GDZ':
                xvec_dependencies['Radius'] = 'R_E'
        return xvec_dependencies

    def create_2Dinterpolator(self, varname, x, y, data, units):
        '''Create gridded interpolator for reconstruction for any six dimension
        pairs.'''

        if x.size == y.size:
            data = data.T  # interpolator gets flipped somehow if square array
        xvec_dependencies = self.xvec_dict()

        # create gridded interpolator, use nearest-neighbor to avoid problems
        # with NaNs.
        rgi = RegularGridInterpolator((x, y), data, bounds_error=False,
                                      fill_value=None, method='nearest')

        '''#generic case
        @kamodofy(units = units, data = data)
        @gridify(x=x, y=y)
        def interpolator_grid(xvec):
            return rgi(xvec)
        '''
        # gridded interpolator inputs must be named according to variables
        # variables not allowed, so need text instead
        coord_list = list(xvec_dependencies.keys())
        if coord_list == ['Time', 'Lon']:
            @kamodofy(units=units, data=data, arg_units=xvec_dependencies)
            @gridify(Time=x, Lon=y)
            def interpolator_grid(thetavec):
                """Interpolates 3d variable into a grid"""
                return rgi(thetavec)

        elif coord_list == ['Time', 'Lat']:
            @kamodofy(units=units, data=data, arg_units=xvec_dependencies)
            @gridify(Time=x, Lat=y)
            def interpolator_grid(thetavec):
                """Interpolates 3d variable into a grid"""
                return rgi(thetavec)

        elif coord_list == ['Time', 'Alt']:
            @kamodofy(units=units, data=data, arg_units=xvec_dependencies)
            @gridify(Time=x, Alt=y)
            def interpolator_grid(thetavec):
                """Interpolates 3d variable into a grid"""
                return rgi(thetavec)

        elif coord_list == ['Time', 'Radius']:
            @kamodofy(units=units, data=data, arg_units=xvec_dependencies)
            @gridify(Time=x, Radius=y)
            def interpolator_grid(thetavec):
                """Interpolates 3d variable into a grid"""
                return rgi(thetavec)

        elif coord_list == ['Time', 'X']:
            @kamodofy(units=units, data=data, arg_units=xvec_dependencies)
            @gridify(Time=x, X=y)
            def interpolator_grid(xvec):
                """Interpolates 3d variable into a grid"""
                return rgi(xvec)

        elif coord_list == ['Time', 'Y']:
            @kamodofy(units=units, data=data, arg_units=xvec_dependencies)
            @gridify(Time=x, Y=y)
            def interpolator_grid(xvec):
                """Interpolates 3d variable into a grid"""
                return rgi(xvec)

        elif coord_list == ['Time', 'Z']:
            @kamodofy(units=units, data=data, arg_units=xvec_dependencies)
            @gridify(Time=x, Z=y)
            def interpolator_grid(xvec):
                """Interpolates 3d variable into a grid"""
                return rgi(xvec)

        elif coord_list == ['Lon', 'Lat']:
            @kamodofy(units=units, data=data, arg_units=xvec_dependencies)
            @gridify(Lon=x, Lat=y)
            def interpolator_grid(thetavec):
                """Interpolates 3d variable into a grid"""
                return rgi(thetavec)

        elif coord_list == ['Lon', 'Alt']:
            @kamodofy(units=units, data=data, arg_units=xvec_dependencies)
            @gridify(Lon=x, Alt=y)
            def interpolator_grid(thetavec):
                """Interpolates 3d variable into a grid"""
                return rgi(thetavec)

        elif coord_list == ['Lon', 'Radius']:
            @kamodofy(units=units, data=data, arg_units=xvec_dependencies)
            @gridify(Lon=x, Radius=y)
            def interpolator_grid(thetavec):
                """Interpolates 3d variable into a grid"""
                return rgi(thetavec)

        elif coord_list == ['Lat', 'Alt']:
            @kamodofy(units=units, data=data, arg_units=xvec_dependencies)
            @gridify(Lat=x, Alt=y)
            def interpolator_grid(thetavec):
                """Interpolates 3d variable into a grid"""
                return rgi(thetavec)

        elif coord_list == ['Lat', 'Radius']:
            @kamodofy(units=units, data=data, arg_units=xvec_dependencies)
            @gridify(Lat=x, Radius=y)
            def interpolator_grid(thetavec):
                """Interpolates 3d variable into a grid"""
                return rgi(thetavec)

        elif coord_list == ['X', 'Y']:
            @kamodofy(units=units, data=data, arg_units=xvec_dependencies)
            @gridify(X=x, Y=y)
            def interpolator_grid(xvec):
                """Interpolates 3d variable into a grid"""
                return rgi(xvec)

        elif coord_list == ['X', 'Z']:
            @kamodofy(units=units, data=data, arg_units=xvec_dependencies)
            @gridify(X=x, Z=y)
            def interpolator_grid(xvec):
                """Interpolates 3d variable into a grid"""
                return rgi(xvec)

        elif coord_list == ['Y', 'Z']:
            @kamodofy(units=units, data=data, arg_units=xvec_dependencies)
            @gridify(Y=x, Z=y)
            def interpolator_grid(xvec):
                """Interpolates 3d variable into a grid"""
                return rgi(xvec)
        else:
            print(coord_list)
            raise AttributeError('Coordinate combination not supported.')

        # store interpolator, units, data, and xvec dictionary and return
        self[varname] = interpolator_grid
        self.variables[varname] = dict(units=units, data=data,
                                       arg_units=xvec_dependencies)
        self.variables[varname]['xvec'] = xvec_dependencies
        return
